# -*- coding: utf-8 -*-
"""
A module provide ability to be compatible with Python 2.7

@author: yanglei
"""

from __future__ import unicode_literals
from __future__ import print_function

from .ylsys import py2, pyi
import os, sys
from functools import reduce, wraps

if py2:
    import __builtin__ as builtins
else:
    import builtins
    
printf = print


if py2:
    unicode = __builtins__['unicode']
else:
    unicode = str

def isstr(s):
    '''
    `isinstance(s, str)` for compatibility both python 2/3
    '''
    return isinstance(s, (str, unicode))


class Classobj:
    ''' 兼容 python 2 的 classobj'''
    pass
classobj = type(Classobj)
def istype(objOrType):
    ''' 
    `isinstance(objOrType, type)` for compatibility both python 2/3
    '''
    return isinstance(objOrType, (type, classobj))


def setFunction2ForAutoList():
    '''
    add function 'range2', 'map2', 'reduce2', 'zip2' that retrun a list like python2
    '''
    funForAutoList = ['range', 'map', 'reduce', 'filter', 'zip', 'enumerate']
    if py2:
        addListAfter = lambda x:x
    else:    
        def addListAfter(f):
            @wraps(f)
            def innerF(*l, **kv):
                r = f(*l, **kv)
                return list(r)
            return innerF
    reduce
    glob = globals()
    for name in funForAutoList:
        rawf = glob[name] if name in glob else  __builtins__[name]
        glob[name+'2'] = addListAfter(rawf)

setFunction2ForAutoList()


def __setDisplayEnv():
    msg = '''%s
        os.environ["DISPLAY"] not found, may cuse error like this 
        [QXcbConnection Error](https://github.com/ipython/ipython/issues/10627)
        so, we auto set os.environ["QT_QPA_PLATFORM"] = "offscreen"    '''%'\x1b[36m%s\x1b[0m'% 'warning from boxx'
    msg = '''os.environ["DISPLAY"] is not found
    plt.show() are redirect to plt.savefig(tmp)
    function: show, loga, crun, heatmap, plot will be affected'''
    
    from .tool import warn
    warn(msg)
    os.environ["QT_QPA_PLATFORM"] = "offscreen"
    os.environ['DISPLAY'] = ':0'
    
def __noDisplayEnv():
    __setDisplayEnv()
    import matplotlib.pyplot as plt
    def savefig(*l, **kv):
        from .ylsys import tmpboxx
        from .tool import increase, warn, OffScreenWarning
        showtmp = os.path.join(tmpboxx(), 'showtmp')
        if not os.path.isdir(showtmp):
            os.makedirs(showtmp, exist_ok=True)
        png = os.path.join(showtmp,'%d_show.png'%increase('showtmp'))
        warn('''os.environ["DISPLAY"] is not found
    plt.show() are redirect to plt.savefig("%s")'''%png, OffScreenWarning)
        plt.savefig(png)
        plt.cla()
        plt.clf()
#    pltshow = plt.show
    plt.show = savefig
def beforImportPlt():
    '''
    cause `import matplotlib.pyplot` often lead to Exit Error
    I do `import plt` when function need, this would avoid exit error.
    if os.environ["DISPLAY"] is not found, then  we auto set os.environ["QT_QPA_PLATFORM"] = "offscreen"
    and plt.show() are redirect to plt.savefig('/tmp/boxxTmp/showtmp')
    
    Farther more, if pyi.gui is False, the plt.show() are replace to plt.savefig('/tmp/boxxTmp/showtmp')
    '''
    if not pyi.plt and not pyi.reloadplt:
        __noDisplayEnv()
        import matplotlib.pyplot as plt
#        import seaborn
        pyi.reloadplt = plt

class SetPltActivateInWith():
    '''
    set plt to interactivate mode while use plt.show() inner boxx
    '''
    def __enter__(self):
        if pyi.gui or not pyi.plt:
            return 
        import matplotlib.pyplot as plt
        self.interactivePlot = plt.rcParams['interactive']
        if not self.interactivePlot and pyi.interactive:
            plt.ion()
    def __exit__(self, *l):
        if pyi.gui or not pyi.plt:
            return 
        import matplotlib.pyplot as plt
        if not self.interactivePlot and pyi.interactive:
            plt.ioff()
    
def interactivePlot(fun):
    '''
    turn interactive plot mode if enviroment is interactive
    '''
    @wraps(fun)
    def innerf(*l, **kv):
        beforImportPlt()
        with SetPltActivateInWith():
            r = fun(*l, **kv)
        return r
    return innerf

if py2:
    execfile = builtins.execfile
else:
    def execfile(filename, globals=None, locals=None):
        '''
        same usage as python2 execfile
        '''
        frame = sys._getframe(1)
        if globals is None:
            globals = frame.f_globals
        if locals is None:
            locals = frame.f_locals
        with open(filename) as f:
            code = f.read()
        exec(code, globals, locals)

def runpyfile(filename, main=True, globals=None, locals=None):
    '''
    add __file__ and __name__ to globals then runpyfile
    like runfile in spyder 
    '''
    default = {
            '__file__':os.path.abspath(filename),
            '__name__':main and '__main__',
            }
    if globals is None:
        globals = {}
    if locals is None:
        locals = {}
    default.update(globals)
    with open(filename) as f:
        code = f.read()
    exec(code, default, locals)

try:
    ModuleNotFoundError = ModuleNotFoundError
except NameError:
    ModuleNotFoundError = ImportError

if not py2 and 0:
    __rawOpen__ = open
    open = lambda *l:__rawOpen__(l[0],'r',-1,'utf8') if len(l) == 1 else __rawOpen__(l[0],l[1],-1,'utf8')




# -*- coding: utf-8 -*-

from __future__ import unicode_literals

'''
Tool-box for efficient build and debug in Python. 
Espacially for Scientific Computing and Computer Vision.
'''
__version__ = "0.10.13"
__short_description__ = "Tool-box for efficient build and debug in Python. Especially for Scientific Computing and Computer Vision."
__license__ = "MIT"
__author__ = "DIYer22"
__author_email__ = "ylxx@live.com"
__maintainer__ = "DIYer22"
__maintainer_email__ = "ylxx@live.com"
__github_username__ = "DIYer22"
__github_url__ = "https://github.com/DIYer22/boxx"
__support__ = "https://github.com/DIYer22/boxx/issues"

from . import ylsys
from . import ylcompat
from . import tool
from . import ylimg
from . import ylsci
from . import ylml
from . import yldb
from . import undetermined



from .tool import *
from .ylsys import *
from .ylcompat import *
from .ylimg import *
from .ylml import *
from .ylsci import *
from .yldb import *

if __name__ == '__main__':
    pass




# -*- coding: utf-8 -*-
"""
tools for detection and bounding box

@author: yl
"""
import numpy as np

from ..ylsci import Vector, cos, sin
from ..tool import sliceInt, dicto
from ..ylimg import toPng, padding

def saveCoco(jsp, cocoDic, imgdf=None, anndf=None, catdf=None):
    '''
    save coco format annotation json according to 3 pandas.DataFrame
    
    Parameters
    ----------
    jsp : string 
        string: path for coco format annaotation json
    cocoDic : dict
        coco format dict
    **df : pandas.DataFrame
        replace the corresponding table in cocoDic 
    '''
    from .. import df2dicts
    cocoDic = cocoDic.copy()
    if imgdf is not None:
        cocoDic['images'] = df2dicts(imgdf)
    if anndf is not None:
        cocoDic['annotations'] = df2dicts(anndf)
    if catdf is not None:
        cocoDic['categories'] = df2dicts(catdf)
    import json
    with open(jsp, 'w') as f:
        json.dump(cocoDic, f)
    return jsp

def loadCoco(jsp):
    '''
    load coco format annotation json to 3 pandas.DataFrame
    
    >>> cocoDic, imgdf, anndf, catdf = loadCoco(jsp)
    
    Parameters
    ----------
    jsp : string or dict
        string: path for coco format annaotation json
        dict: a coco format json
    Return
    ----------
    cocoDic, imgdf, anndf, catdf
    '''
    import json
    import pandas as pd
    if not isinstance(jsp, dict):
        with open(jsp, 'r') as f:
            cocoDic = json.load(f)
    else:
        cocoDic = jsp
    imgdf = pd.DataFrame(cocoDic['images']).set_index('file_name', False)
    anndf = pd.DataFrame(cocoDic['annotations'])
    catdf = pd.DataFrame(cocoDic['categories']).set_index('id', False)
    return cocoDic, imgdf, anndf, catdf

#inttuple = mf - inttuple
def cropMinAreaRect(img, rect, borderValue=None):
    import cv2
    # rotate img
    rect = [(int(rect[0][0]),int(rect[0][1])),
            (int(rect[1][0]),int(rect[1][1])),
            int(rect[2])]
    angle = rect[2]
    rows,cols = img.shape[0], img.shape[1]
    M = cv2.getRotationMatrix2D((cols/2,rows/2),angle,1)
    img_rot = cv2.warpAffine(img,M,(cols,rows),borderValue=borderValue)
    
    # rotate bounding box
#    rect0 = (rect[0], rect[1], 0.0)
    box = cv2.boxPoints(rect)
    pts = np.int0(cv2.transform(np.array([box]), M))[0]    
    pts[pts < 0] = 0

    # crop
    img_crop = img_rot[pts[1][1]:pts[0][1], 
                       pts[1][0]:pts[2][0]]
    return img_crop

    if 0 == 'test':
        # generate image
        img = np.zeros((1000, 1000), dtype=np.uint8)
        img = cv2.line(img,(400,400),(511,511),1,120)
        img = cv2.line(img,(300,300),(700,500),1,120)
        img = cv2.line(img,(0,0),(1000,100)[::-1],2,200)
        
        # find contours / rectangle
        _,contours,_ = cv2.findContours(img, 1, 1)
        
        # rect = ((middle_point), (w, h), angle)
        rect = cv2.minAreaRect(contours[0])
    #    rect = ((100.0, 499.5), (1999.0, 200.0), -90.0)
        print(rect)
        from boxx import timeit, show
        # crop
        with timeit():
            img_croped = cropMinAreaRect(img, rect)
        show(img)
        show(img_croped)
    
class Bbox():
    '''
    
    sbox:
        小方块 ,1. 减少旋转的计算量;2. 方便贴图
    urdl:
        mean up, right, down, left
    expandPara:
        face bbox to head bbox, evalue by Bbox.getExpandFun(imgBox, sbox, rate=256/286.)
    vx, vy:
        vx: w 方向的向量; vy: h 方向的向量
    ps, p0:
        ps: bbox 的四个定点, p0 左上角为 p0 顺时针依次类推
    
    '''
    def __init__(self, p0, w, h, deg=None, canvas=None, expandPara=None, **kv):
        vx, vy = Bbox.getVxVy(w, h, deg)
        pNose = p0 + vx/2 + vy/2
        if expandPara :
            unit = (w + h)/2
            delta = unit * expandPara.vRate
            if deg:
                delta = delta.rotation(deg)
            p0 = delta + pNose
            w = unit * expandPara.wRate
            h = unit * expandPara.hRate
            vx, vy = Bbox.getVxVy(w, h, deg)
        ps = np.array([p0, p0+vx, p0+vx+vy, p0+vy])
        
        urdl = u, r, d, l = [min(ps[:, 1]), max(ps[:, 0]), max(ps[:, 1]), min(ps[:, 0])]
        self.offsetToSbox = Vector([l, u])
        self.boxUrdl = None
        self.__dict__.update(locals())
    def __str__(self, ):
        p0, w, h, deg = self.p0, self.w, self.h, self.deg
        return "p0:{p0}, w:{w}, h:{h}, deg:{deg}".format(p0=p0, w=w, h=h, deg=deg)
    __repr__ = __str__
    
    @staticmethod
    def getVxVy(w, h, deg=None):
        if deg is None:
            vx, vy = Vector([w, 0]), Vector([0, h])
        else:
            sinn, coss = sin(deg), cos(deg)
            vx = Vector([coss*w, sinn*w])
            vy = Vector([-sinn*h, coss*h])
        return vx, vy

#        def expand(d_urdl=[98, 44, 42, 44]):
#            178 * 218

#    @staticmethod
    def _getUrdlForCanvas(self):
        '''get sbox urdl in canvas
        '''
        if self.boxUrdl is None: 
            dx, dy = self.canvas.offset
            u, r, d, l = self.urdl
            self.boxUrdl = u+dy, r+dx, d+dy, l+dx
        return self.boxUrdl
    def cropSbox(self):
        u, r, d, l = self._getUrdlForCanvas()
        sboxImg = self.canvas.canvas[sliceInt[u:d, l:r]]
        sboxImgPng = toPng(sboxImg)
        return sboxImgPng
    def paste(self, rimg):
        '''
        rimg: result img'''
        u, r, d, l = self.u, self.r, self.d, self.l 
        if self.deg is None:
            mask = slice(None)
            if rimg.shape[-1] == 4:
                rimg, mask = rimg[...,:3], rimg[...,3] > 0 
            self.canvas.canvas[sliceInt[u:d, l:r]][mask] = rimg[mask]
            return 
        rimg = toPng(rimg)
        w, h = r-l, d-u
        deg = - self.deg
        p2 = Vector((rimg).shape)[[1,0]]
        pad = max([w,h]-p2)/2
        padedImg = padding(rimg, pad)
        rect = (p2/2+pad, (w, h), deg)
        simgPng = cropMinAreaRect(padedImg, rect, borderValue=(0,0,0,0))
        simg, mask = simgPng[...,:3], simgPng[...,3]>simgPng[...,3].max()/2
        u, r, d, l = self._getUrdlForCanvas()
#        g()
#        self.canvas.canvas[sliceInt[u:d, l:r]][mask] = simg[mask]
        self.canvas.canvas[sliceInt[u:u+mask.shape[0], l:l+mask.shape[1]]][mask] = simg[mask]
        return self.canvas.canvas
    def crop(self, cropSboxFirst=False):
        if cropSboxFirst :
            img = self.cropSbox()
            p0 = self.p0 - self.offsetToSbox
        else:
            img = self.canvas.canvas
            p0 = self.p0
        w, h, deg = self.w, self.h, self.deg
        l, u = p0
        vx, vy = self.vx, self.vy
        if deg is None:
            deg = 0
        
        pm = p0 + vx/2. + vy/2.
        rect = (tuple(pm.tolist()), (w.tolist(), h.tolist()), deg)
        img_croped = cropMinAreaRect(img, rect, borderValue=(0,0,0,0))
        return (img_croped)
    @staticmethod
    def test():
        pass
    

class Canvas():
    '''
    a canvas to padding bound, draw bbox, paste new patch
    raw, img:
        原始图像
    canvas: 
        根据扩张后的bbox padding 的画布
    offset:
        canvas 和 raw 的坐标转换: v_canvas = v_raw + offset
    '''
    BboxClass = Bbox
    def __init__(self, img, bboxList=None, expandPara=None):
        if bboxList is None:
            bboxList = []
        self.raw = img
        self.bboxs = bboxs = []
        for bboxDic in bboxList:
            dic = dicto(canvas=self, expandPara=expandPara,)
            if isinstance(bboxDic, Bbox):
                bbox = bboxDic
                bbox.__init__(bbox.p0, bbox.w, bbox.h, deg=bbox.deg, **dic)
            else:
                dic.update(dict(bboxDic))
                bbox = self.BboxClass( **dic)
#            g()
            bboxs.append(bbox)
        h, w = img.shape[:2]
        self.urdl = np.array([[0, w, h, 0]] + [b.urdl for b in bboxs])
        self.offset = dx, dy = -Vector(self.urdl[:, [3, 0]].min(0), )
        self.leftDown = self.urdl[:, [1,2]].max(0)
        ww, hh = (self.offset + self.leftDown).intround()
        self.canvas = np.zeros((hh, ww,) + img.shape[2:], img.dtype)
        self.canvas[sliceInt[dy:dy+h, dx:dx+w]] = img
#        self.bias = Vector([0, 0])
        self.__dict__.update(locals())
#        g()

    def getResult(self):
        dy, dx, h, w = self.dy, self.dx, self.h, self.w
        return self.canvas[sliceInt[dy:dy+h, dx:dx+w]]
    def show(self):
        from boxx import show, pblue
        canvas = self
        result = canvas.getResult()
        pblue('show canvas.getResult()')
        show-result
        pblue('show Bboxs.crop(): crop from canvas.raw')
        show-[b.crop() for b in canvas.bboxs]
        if canvas.bboxs and canvas.bboxs[0].deg:
            pblue('show Bboxs.crop(cropSboxFirst=True):')
            show-[b.crop(cropSboxFirst=True) for b in canvas.bboxs]
            pblue('show Bboxs.cropSbox():')
            show-[b.cropSbox() for b in canvas.bboxs]
    
if __name__ == "__main__":
    pass
    
    



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: DIYer22@github
@mail: ylxx@live.com
Created on Fri Jun 21 13:38:11 2019
"""
from __future__ import unicode_literals
import os

from .toolStructObj import FunAddMagicMethod

def startGnomeTerminal(dirr=None):
    from ..ylsys import linuxYl
    assert linuxYl
    if dirr is None:
        os.system('gnome-terminal ')
    elif os.path.isfile(dirr):
        os.system('gnome-terminal  --working-directory "%s"'%os.path.abspath(os.path.dirname(dirr)))
    else:
        os.system('gnome-terminal  --working-directory "%s"'%os.path.abspath(dirr))
        
ter = FunAddMagicMethod(startGnomeTerminal)


def startNautilus(dirr=None):
    from ..ylsys import linuxYl
    assert linuxYl
    if dirr is None:
        os.system('nautilus .')
    elif os.path.isfile(dirr):
        os.system('nautilus "%s"'%os.path.dirname(dirr))
    else:
        os.system('nautilus "%s"'%dirr)
nau = FunAddMagicMethod(startNautilus)


def startVscode(dirr=None, relpath=None):
    if dirr is None:
        dirr = '.'
    dirr = os.path.abspath(dirr)
    if relpath is not None:
        dirr = os.path.abspath(os.path.join(dirr, relpath))
    os.system('code "%s"' % dirr)
    
vscode = startVscode

if __name__ == "__main__":
    pass
    
    
    



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: DIYer22@github
@mail: ylxx@live.com
Created on Fri Feb 15 20:20:59 2019
"""
from . import torch, hasnan


def pthnan(pth):
    dic = torch.load(pth, map_location="cpu")
    from collections import OrderedDict

    def getOrderedDict(seq):
        if isinstance(seq, OrderedDict):
            return seq
        if isinstance(seq, dict):
            seq = list(seq.values())
        if not isinstance(seq, (tuple, list)):
            return None
        for s in seq:
            re = getOrderedDict(s)
            if re is not None:
                return re

    od = getOrderedDict(dic)
    tensor = list(od.values())[-1]
    print(
        '\n"%s"\n\nHas nan: %s\n'
        % ("\x1b[36m%s\x1b[0m" % tensor[..., :10], "\x1b[31m%s\x1b[0m" % hasnan(tensor))
    )


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(
        description="""
    detect a pth file "Does it has nan?" """
    )
    parser.add_argument(
        "pth",
        default="/home/dl/github/maskrcnn/output/mix_11/model_final.pth",
        type=str,
    )
    args = parser.parse_args()

    pthnan(pth=args.pth)



# -*- coding: utf-8 -*-
"""
tools for Machine Learning and Image Processing

@author: yanglei
"""
from __future__ import unicode_literals

# from .ylmlTrain import GenSimg

# from .ylmlTest import (binaryDiff, classDiff, drawBoundAndBackground, 
#                       confusionMatrix, f1Score,
#                       getWeightCore, smallImg, autoSegmentWholeImg,
#                       ArgList, autoFindBestEpoch, autoFindBestParams)

# from .ylmlEvalu import (Evalu, accEvalu, lplrEvalu, diceEvalu, pd)

from .ylDete import Bbox, Canvas, cropMinAreaRect, loadCoco, saveCoco

if __name__ == "__main__":
    pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: DIYer22@github
@mail: ylxx@live.com
Created on Mon Mar  2 12:52:18 2020

Make python script as easy to use as shell script.

Usage:
    python -m boxx.script '[print(i) for i in range(10)]'
    
    # add "tmp_" as prefix for all file in '.' 
    python -m boxx.script '[mv(fname, "tmp_"+fname) for fname in ls()]'

alias "ls, mv, ln, lns, cp, cpr" as funcation.

"""
import boxx
from os import *
from os.path import *
from boxx import *
from builtins import *


_tmp_list_as_None = []
class CacheVar(dict):
    def __init__(self, name="cache"):
        super().__init__()
        self.__mxscript_cache_call_value__ = _tmp_list_as_None
        self.__mxscript_cache_name__ = name


    def __setattr__(self, name, value):
        self[name] = value
        return value

    set = __setattr__
    
    def __getattr__(self, name):
        if name in self:
            return self[name]
        return super().__getattribute__(name)
    
    def __call__(self, value=_tmp_list_as_None):
        """Store one tmp var by call
        >>> print(tmp('tmp_value1'), tmp(), tmp('tmp_value2'), tmp())
        tmp_value1 tmp_value1 tmp_value2 tmp_value2
        """
        if value is _tmp_list_as_None:
            if self.__mxscript_cache_call_value__ is _tmp_list_as_None:
                raise ValueError(f"Please register {self.__mxscript_cache_name__} value by `{self.__mxscript_cache_name__}(value)` first!")
            return self.__mxscript_cache_call_value__
        else:
            self.__mxscript_cache_call_value__ = value
            return value

d = CacheVar("d")
tmp = CacheVar("tmp")
cache = CacheVar("cache")
cache1 = CacheVar("cache1")
cache2 = CacheVar("cache2")
cache3 = CacheVar("cache3")


import os
import sys
import shutil
from os import rename, link, symlink, remove

mv = rename
rm = remove
ln = link
lns = symlink
copy = shutil.copy
copytree = shutil.copytree

def sh(cmd):
    return boxx.execmd(cmd).strip().split()
bash = sh

def ls(path=".", a=False):
    cmd = 'ls "' + path + '"' + (' -a ' if a else '')
    return sh(cmd)

def cp(src, dst):
    cmd = 'cp -r "' if os.path.isdir(src) else 'cp "'
    cmd += src + '" "' + dst + '"'
    return sh(cmd)

cpr = cp
# TODO cp cpr support oss

def cat(*paths):
    for path in paths:
        if len(paths) > 1:
            print("-"*5, path, "-"*5)
        with open(path, 'r') as f:
            print(f.read())

builtins_dict = __builtins__ if isinstance(__builtins__, dict) else __builtins__.__dict__
class GetKey(dict):
    def __getitem__(self, k):
        if k in self:
            return dict.__getitem__(self, k)
        elif k in builtins_dict:
            return builtins_dict[k]
        else:
            try:
                print('try import %s' % (k,))
                self[k] = __import__(k)
            except ModuleNotFoundError as e:
                if e.name != k:
                    raise e
            return dict.__getitem__(self, k)


context = GetKey(globals())

def main():
    if len(sys.argv) <= 1:
        __file__ = os.path.abspath('temp.py')

        import IPython

        IPython.embed()
    else:
        code = " ".join(sys.argv[1:])
        if ";\\n" in code: # support multi line
            code = code.replace(';\\n', ';\n')
            print('Code: """\n%s\n"""' % code)
        else:
            print('Code: "%s"' % code)
        print()
    
        exec(code, context, context)

if __name__ == "__main__":
    main()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
'''
`import boxx.g` act same as `g()`

useage:
    >>> import boxx.g
        
Instruct
----------
    your can exec `g()`, `p()` in any function or module
    all vars that belong the function will transport to
    Python interactive shell. and `globals()` will in `p` which is a dicto
    the differtent between `g` and `gg` is that `gg` will auto print the vars's information
    
    BTW, `import boxx.g`, `import boxx.p` is convenient way to use `g()` without `from boxx import g`
    
Notice
-----------
    For Python 2 only first time import works 

    在函数内运行`p()` or `lc()`  
    则此函数的global和local 变量会载入全局变量 p 中
    函数的 frame等其他信息 则放入全局变量 lc

'''
from .ylsys import  py2
from .tool import out

deep = 6
if py2:
    deep = 1

out(deep, printt=False)

if py2:
    import sys
    from .tool import g
    sys.modules['boxx.g'] = g
else:
    from .tool import removeImportSelf
    removeImportSelf('boxx.g')


if __name__ == '__main__':
    pass



# -*- coding: utf-8 -*-
"""
Math and Scientific Computing module

@author: yl
"""

from __future__ import unicode_literals

from .ylnp import  e, pi, nan, inf, eps
from .ylnp import getNumpyType, mapping_array
from .ylnp import savenp, loadnp, plot3dSurface, isNumpyType, testNumpyMultiprocessing

from .ylvector import sin, cos, tan, arcsin, arccos, arctan, deg2rad, rad2deg
from .ylvector import Vector, v0

from .ylstatistics import distavg, distnorm, DiscreteSample, HeatmapSample


# -*- coding: utf-8 -*-

from __future__ import unicode_literals
'''
Todo List:
    1. 使用ORM 映射
'''
import sqlite3,os

class Sqlite():
    def __init__(self, creatSql, dbpath='sqlite3.db'):
        '''
        创建一个sqlite3 管理器
        creatSql: 构建数据库代码
        dbPath: 数据库路径
        '''
        self.creatSql = creatSql
        self.dbpath = dbpath
        self.creatDbIfNot()
    def creatDbIfNot(self):
        '''
        注意：最后一句不能以逗号结尾
        '''
        dbpath = self.dbpath
        if not os.path.isfile(dbpath):
            dirr = os.path.dirname(dbpath)
            if not os.path.isdir(dirr) and dirr:
                os.makedirs(dirr, exist_ok=True)
            self.exe(self.creatSql)
        
    def exe(self, sql, values=None):
        with sqlite3.connect(self.dbpath) as conn:
            cursor = conn.cursor() 
            if values is None :
                cursor.execute(sql)
            else:
                cursor.execute(sql,values)
            cursor.close()
            conn.commit()
    def insert(self, sql, value):
        self.inserts(sql, [value])
    def inserts(self, sql, values):
        '''
        一次性插入多条语句再commit性能好
        '''
        with sqlite3.connect(self.dbpath) as conn:
            cursor = conn.cursor() 
            for value in values:
                if not isinstance(value,(tuple,list)):
                    value = (value,)
                cursor.execute(sql,value)
            cursor.close()
            conn.commit()
    def select(self, sql, values=None):
        '''
        执行select 返回结果
        '''
        dbpath = self.dbpath
        with sqlite3.connect(dbpath) as conn:
            cursor = conn.cursor() 
            conn = sqlite3.connect(dbpath)
            if values is None :
                cursor.execute(sql)
            else:
                cursor.execute(sql,values)
            resoult = cursor.fetchall()
            cursor.close()
        return resoult
        
if __name__ == '__main__':
    import random as rd
    creat = '''CREATE TABLE t(
    no INTEGER PRIMARY KEY ,
    id INTEGER NOT NULL
);'''
    s = Sqlite(creat)
    insert = 'insert into t (id) values (?)'
    v = rd.randint(0,100)
    s.insert(insert,v)
    
    select = 'select * from t'
    r = s.select(select)
    print(v,r)
    
    pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
statistics tools 

@author: yl
"""
import random
import numpy as np
from functools import wraps
from collections import Counter


def Distribution(f):
    """
    return  distribution of bias round v and max bias is @maxratio*v.mean() or @maxbias
    """

    @wraps(f)
    def innerF(v, maxbias=None, maxratio=0.2, *l, **kv):
        mean = v
        if maxbias is None:
            if isinstance(v, np.ndarray):
                mean = v.mean()
            maxbias = maxratio * mean
        biass = f(v=v, maxbias=maxbias, maxratio=maxratio, *l, **kv)
        return v + biass

    return innerF


@Distribution
def distnorm(v, maxbias=None, maxratio=0.2, std=2):
    """
    return normal distribution of bias round v and max bias is @maxratio*v.mean() or @maxbias
    """
    shape = None
    if isinstance(v, np.ndarray):
        shape = v.shape
    normaBiass = np.random.normal(loc=1, scale=1.0 / std, size=shape) % (2) - 1

    biass = normaBiass * maxbias
    return biass


@Distribution
def distavg(v, maxbias=None, maxratio=0.2):
    """
    return uniform distribution of bias round v and max bias is @maxratio*v.mean() or @maxbias
    """
    shape = None
    if isinstance(v, np.ndarray):
        shape = v.shape
    biass = np.random.uniform(-maxbias, maxbias, size=shape)
    return biass


class DiscreteSample(dict):
    def sample(self):
        summ = sum(self.values())
        random_number = random.random() * summ
        for k, v in self.items():
            random_number -= v
            if random_number <= 0:
                return k

    @staticmethod
    def test():
        ds = DiscreteSample({"a": 1, "b": 2, "c": 0})
        print(Counter([ds.sample() for _ in range(1000000)]))


class HeatmapSample:
    def __init__(self, heatmap):
        h, w = heatmap.shape
        hs, ws = np.mgrid[:h, :w]
        hs, ws = hs.reshape(-1), ws.reshape(-1)
        self.discrete_sample = DiscreteSample(zip(zip(hs, ws), heatmap.reshape(-1)))

    def sample(self):
        return self.discrete_sample.sample()


if __name__ == "__main__":
    from boxx import loga

    loga(distavg(np.ones((517, 111))))
    loga(distnorm(np.ones((517, 111)), 3, std=4))
    DiscreteSample.test()
    pass



# -*- coding: utf-8 -*-

from __future__ import unicode_literals
from __future__ import print_function

from .toolStructObj import addCall, dicto, FunAddMagicMethod, getfathers, typeNameOf, typestr
from .toolSystem import getRootFrame, getFatherFrames
from ..ylsys import py2
from ..ylcompat import printf, unicode
from ..ylcompat import istype

import os,sys,time
import math
import re
from collections import defaultdict

def _str_struct_time(t, st, int_sec=False):
    s = "%d-%02d-%02d-%02d_%02d_%02d"%(st[:6])
    if int_sec:
        return s
    s = s + ("%.5f"%(t%1))[1:]
    return s

def localTimeStr(int_sec=False, asctime=False):
    '''
    get local time in string format like '2020y_02m_26d_04h_05m_41.38258s'
    获得本地时间(GM+8 北京时间)
    '''
    t = time.time()
    local_struct_time = time.localtime(t)
    
    if asctime:
        return time.asctime(local_struct_time)
    return _str_struct_time(t, local_struct_time, int_sec)

def gmtTimeStr(asctime=False):
    '''
    get GMT(global) time in string format like '2020y_02m_26d_04h_05m_41.38258s'
    获得gmt时间(GM0 格林威治时间)
    '''
    t = time.time()
    gmt_struct_time = time.gmtime(t)
    if asctime:
        return time.asctime(gmt_struct_time)
    return _str_struct_time(t, gmt_struct_time)

def time_str_to_stamp(time_str):
    new_time_str = time_str.replace("-", "_")[:19]
    stamp = time.mktime(time.strptime(new_time_str, "%Y_%m_%d_%H_%M_%S"))
    if '.' in time_str:
        stamp += float(time_str[19:])
    return stamp

class timeGap:
    '''
    定时log器 隔固定的一段时间 返回 当前轮数(True)
    
    Init Parameters
    ----------
    gap : float
        隔多少秒返回一个 True
    fun : function, default None
        每一次要执行fun()
    quickBegin : bool, default True
        First time call will return True
    
    Attribute
    ----------
    __call__ : 
        返还 当前轮数(True)
    time : @property
        is @property 返回已记录的时长
    '''
    def __init__(self, gap=1,fun=False, quickBegin=True):
        self.n = 0
        self.gap = gap
        self.begin = time.time()
        self.last = self.begin - (quickBegin and gap)
        self.fun = fun
    def __call__(self,):
        t = time.time()
        if t - self.last >= self.gap:
            self.last = t
            self.n += 1
            if self.fun:
                self.fun()
            return self.n
        return False
    @property
    def time(self):
        return time.time() - self.begin

TimeGapDic = {}
def timegap(gap=10, key='boxx.default', quickBegin=True):
    '''
    定时器 隔固定的一段时间 返回 当前轮数(True), use to log in loop
    This is light version of timeGap, for More Infomation help(timeGap)
    
    Init Parameters
    ----------
    gap : number
        隔多少秒返回一个 True
    key : hashable, default 'boxx.default'
        the namespace to count
    '''
    keyt = (gap, key)
    if keyt not in TimeGapDic:
        TimeGapDic[keyt] = timeGap(gap, quickBegin=quickBegin)
    return TimeGapDic[keyt]()

__logGapDic = {}
def logGap(key='gap'):
    '''
    log time gap in loop
    '''
    d = __logGapDic
    l = d.get(key)
    now = time.time()
    if not l :
        pblue('Begin to log time gap of "%s"!'%key)
        d[key] = [0, now]
    else :
        dt = now - l[1]
        pblue('%dst loop, "%s" spend %ss.'%(l[0] ,key, strnum(dt)))
        l[0] += 1
        l[1] = now

frontColorDic = dicto({   # 前景色
        'black'    : 30,   #  黑色
        'red'      : 31,   #  红色
        'green'    : 32,   #  绿色
        'yellow'   : 33,   #  黄色
        'blue'     : 34,   #  蓝色
        'purple'   : 35,   #  紫红色
        'cyan'     : 36,   #  青蓝色
        'white'    : 37,   #  白色
    })
class colorFormat:
    '''
    同时兼容windows和Linux及各类明暗色调的颜色str temple
    只有 red, cyan(==blue), purple
    使用colorFormat.printAllColor()查看效果
    用法: print colorFormat.blue%strr
    '''
    red = '\x1b['+str(frontColorDic['red'])+'m%s\x1b[0m'
    cyan= '\x1b['+str(frontColorDic['cyan'])+'m%s\x1b[0m'
    purple = '\x1b['+str(frontColorDic['purple'])+'m%s\x1b[0m'
    black = '\x1b['+str(frontColorDic['black'])+'m%s\x1b[0m'
    blue = cyan
    info = blue
    danger = red
    err = red
    b = blue
    r = red
    p = purple
    @staticmethod
    def printAllColor(s='printAllColor'):
        return [stdout((c+' '*10)[:7]) and pcolor(c,": "+str(s)) for c in frontColorDic] and None
    pall = printAllColor
clf = colorFormat

def decolor(colored):
    '''
    remove color of str
    '''
    pa = re.compile('\x1b\[3[0-9]m|\x1b\[0m')
    new = pa.sub('', colored)
    new = pa.sub('', new)
    return new

def tounicode(strr):
    '''
    Python2 中文编码错误解决方案，用于代替str类及函数, 全部换成unicode
    '''
    if not py2:
        if isinstance(strr,str):
            return strr
        if isinstance(strr,bytes):
            return strr.decode('utf-8','replace')
        return str(strr)
    else:
        if isinstance(strr,unicode):
            return strr
        strr = str(strr)
        if isinstance(strr,str):
            return strr.decode('utf-8','replace')
        return unicode(strr)

def tostrpy2(x):
    '''
    only work py2 return a str to avoid 
    UnicodeEncodeError when `str(u'中文')`
    '''
    if not py2:
        return str(x)
    if isinstance(x,unicode):
        return x.encode('utf-8','replace')
    s = str(x)
    return s

def shortDiscrib(x):
    '''
    short Discrib of anything for logc number is better 
    '''
    if isinstance(x, int):
        return str(x)
    typee = typestr(x)
    from ..ylimg import StructLogFuns
    fund = StructLogFuns
    f = shortStr
    if typee in fund:
        f = fund[typee]
        if 'torch.' in typee and 'ensor' in typee and (not x.shape):
            f = strnum
    elif '__float__' in dir(x):
        try:
            x = float(x)
            f = strnum
        except:
            pass
    try:
        return decolor(f(x))
    except:
        return str((x))

def shortStr(x, maxlen=60):
    '''
    genrate one line discrib str shorter than maxlen.
    if len(s)> maxlen then slice additional str and append '...' 
    BTW. function will replace '\n' to '↳'
    
    if bool(maxlen) is False, not be short
    '''
    s = tounicode(x).strip()
    if maxlen and len(s) > maxlen:
        s = s[:maxlen-3]
        s +=  '...'
        if '\x1b[' in s :
            s += '\x1b[0m' * (list(s).count('\x1b'))
    s = s.replace('\n','↳')
    return s

def discrib(x, maxline=20):
    '''
    return discrib of x less than maxline.
    if len(s)> maxline*80 then slice additional str and append '...' 
    
    if bool(maxline) is False, return s
    '''
    s = tounicode(x)
    if not maxline:
        return s
    enters = s.count('\n')+1
    n  = len(s)
    maxlen = maxline * 80
    if n > maxlen:
        s = s[:maxlen-6]+'......'
    elif enters > maxline:
        s = '\n'.join(s.split('\n')[:maxline]) + '\n......'
    else:
        return s
    if '\x1b[' in s :
        s += '\x1b[0m' * (list(s).count('\x1b'))
    return s


def logc(code, exe=None):
    '''
    short of log code
    pretty print expression by show every var's value in expression
    
    Parameters
    ----------
    code : str
        the expression code that want to print
    exe : bool, default None
        wether exec(code) befor print
        when exe is None:
            try not exec(code)
            but if some var in expression can't find in locals() and globls()
                then exec(code)
    
    TODO:
        1. use re to replace vars name avoid the same names
        2. use Abstract Syntax Tree and re 
           to distinguish .attr, function call []
    '''
    frame = sys._getframe(2)
    local = frame.f_locals
    glob = frame.f_globals
    
    if exe:
        exec(code, local, glob)
    
    varss = re.findall('[a-zA-Z_][a-zA-Z0-9_]*',code)
    dic = {}
    for name in varss:
        if name in local:
            dic[name] = local[name]
        elif name in glob:
            dic[name] = glob[name]
        elif exe is None:
            exec(code, local, glob)
            if name in local:
                dic[name] = local[name]
            else:
                dic[name] = addCall(name)
    coder = numr = code
    
    toNotVarName = lambda name: '$_%s_$'% ''.join([str(ord(c)) for c in name])
    addSpaceBothSide = lambda s, n: (n//2)*' ' + str(s) + (n-n//2)*' '
    
    for k, v in sorted(dic.items(), key=lambda x:-len(x[0])):
        vstr = shortDiscrib(v)
        if callable(v):
            vstr = k
        maxs = max(len(vstr), len(k))
        
        dic[k] = dict(
                vstr=vstr,
                maxs=maxs,
                v=v,
                ks=addSpaceBothSide(k, maxs-len(k)),
                vs=addSpaceBothSide(vstr, maxs-len(vstr)),
                )
        coder = coder.replace(k, toNotVarName(k))
        numr = numr.replace(k, toNotVarName(k))
    
    for k, d in sorted(list(dic.items()), key=lambda x:-len(x[0])):
        coder = coder.replace(toNotVarName(k), clf.b%d['ks'])
        numr = numr.replace(toNotVarName(k), clf.r%d['vs'])
    
    s = ('Code: %s\n  └── %s'%(coder, numr))   
    print(s)
logc = FunAddMagicMethod(logc)

def tabstr(s, head=4, firstline=False):
    '''
    to tab a block of str for pretty print
    
    Parameters
    ----------
    head : str or int, default 4
        the str that to fill the head of each line
        if value is int, head = ' '*head
    firstline : bool, default False
        whether fill the first line 
    '''
    if isinstance(head, int):
        head = ' '*head
    if firstline:
        s = head + s
    return s.replace('\n','\n'+head)

def getDoc(f):
    '''
    get document of f, if f don't have __doc__, return None
    '''
    if '__doc__' in dir(f) and f.__doc__:
        return f.__doc__
    return None
        
def pcolor(color, *s):
    '''
    用颜色打印 不返回
    '''
    if not isinstance(color,int):
        color = frontColorDic[color]
    print(('\x1b[%dm%s\x1b[0m'%(color, ' '.join(map(tounicode,s)))))
    return 
    if len(s) == 1:
        return s[0]
    return s
pcyan = lambda *s:pcolor('cyan',*s) #  青蓝色
pred = lambda *s:pcolor('red',*s) #  红色
ppurple = lambda *s:pcolor('purple',*s) #  紫红色
pblue = pcyan 

def stdout(*l):
    sys.stdout.write(('%s'%(' '.join(map(tounicode,l)))))
    if len(l) == 1:
        return l[0]
    return l

class Log():
    def __call__(self,*l, **kv):
        printf(*l, **kv)
    def __div__(self, x):
        printf(x)
        return x
    __sub__ = __call__
    __truediv__ = __div__
log = FunAddMagicMethod(printf)
printt = log

def printToStr(value='', *l, **kv):
#def printToStr(value='', *l, sep=' ', end='\n', file=None, flush=False): # for py version < 2.7.16
    '''
    same usage to print function, but replace stdout to return str
    '''
    sep=' '
    end='\n'
    file=None
    flush=False
    locals().update(kv)
    l = (value,)+l
    s = sep.join([tounicode(v) for v in l]) + end
    return s

class PrintStrCollect():
    from functools import wraps
    def __init__(self, ):
        self.s = ''
    @wraps(printToStr)
    def __call__(self, *l, **kv):
        s = printToStr(*l, **kv)
        self.s += s
    def __str__(self):
        s = self.s
        return tostrpy2(s)
    pass

class LogAndSaveToList(list):
    '''
    存储最近用于logg的结果, 并返回x的class, list版本SuperG
    '''
    def __init__(self,printFun=None,cache=5):
        self.p = printFun or pblue
        self.cache = cache
    @property
    def i(self):
        return self[-1]
    _ = i
    @property
    def ii(self):
        return self[-2]
    __ = ii
    @property
    def iii(self):
        return self[-3]
    ___ = iii
    @property
    def iiii(self):
        return self[-4]
    ____ = iiii
    def __call__(self,*l,**dic):
        while self.cache < len(self):
            self.remove(self[0])
        if len(l) == 1:
            self.p(l[0])
            self.append(l[0])
            return l[0]
        self.p('\n'.join(map(tounicode,l)))
        if len(dic):
            for k in dic:
                self.p('%s = %s'%(k,tounicode(dic[k])))
            self.append(dic)
        self.append(l)
        return l
    __sub__ = __lshift__ = __rshift__  = __div__ = __truediv__ =__call__
    def __repr__(self):
        blue  = '\x1b[%dm%s\x1b[0m'%(frontColorDic['red'],tounicode(len(self) and self[len(self)-1]))
        return '''LogAndSaveToList(printFun=%s, cache=%s) log[-1]: %s'''%(
        str(self.p),self.cache,blue)
    __str__ = __repr__

logg = LogAndSaveToList()
pcyan,pred,ppurple,stdout = list(map(FunAddMagicMethod,
                                        [pcyan,pred,ppurple,stdout]))
    
pblue = pcyan 
pinfo = pblue
pdanger = pred
perr = pred


def notationScientifique(num, roundn=None, tuple=False):
    '''
    科学计数法 

    roundn : int, default None
        Notation Scientifique
        保留有效位数 
    tuple : bool, default False
        if True, return a tuple(head, pow) instead of string
    '''
    if num == 0:
        head, pow = 0,0
    else:
        loged = (math.log(abs(num),10))
        pow = int(loged) + (loged < 0 and -1)
        head = num*10**-pow
        if roundn is not None:
            head = round(head, roundn-1)
    if tuple:
        return (head, pow)
    s = '%se%d'%(str(head),pow)
    return s

def strnum(num, roundn=4):
    '''
    better str(num) avoid to long round
    support nan inf
    '''
    try:
        if isinstance(num, int):
            return str(num)
        elif not isinstance(num, float) and '__float__' in dir(num):
            num = float(num)
        head, pow = notationScientifique(num, roundn=roundn, tuple=True)
        if pow > roundn or pow < -min(3,roundn):
            s = '%se%d'%(str(head),pow)
        else:
            s = str(round(num, -pow+roundn))
        return s
    except (OverflowError, ValueError) as e:
        import numpy as np
        if np.isnan(num) or np.isinf(num):
            return str(num)
        raise e

def percentStr(num, roundn=2, sign=False):
    '''
    float to percent string
    roundn mean round to percent
    '''
    num = round(num*100, roundn)
    tabn = (2 if num <10 else 1 )if num <100 else 0
    s = (' '*tabn+'%.'+str(roundn)+'f')%(num)
    if sign:
        s += "%"
    return s
               
def ignoreWarning():
    from warnings import filterwarnings
    filterwarnings('ignore')


class LogLoopTime():
    '''
    记录循环信息的对象 主要的信息有每个循环时间和进度
    names: 用于循环的list，tuple
    digits: 时间的保留位数 可为负数
    loged: 是否直接打印
    
    使用时候 在循环最后加上 logLoopTime(name) 即可
    '''
    def __init__(self,names=None,digits=3,loged=True):
        self.ns = names
        self.begin = self.t = time.time()
        self.digits = digits
        self.loged = loged
        self.count = 0
    def __call__(self,name):
        digits = self.digits
        t = time.time()
        dt = t - self.t
        self.t = t
        times = ('%.'+str(digits)+'f')%dt if digits > 0 else '%.2fE%d'%(dt/10**abs(digits-1),abs(digits-1))
        if self.ns is None:
            s = 'No.%d  %s time:%s'%(self.count,name, times)
            self.count += 1
        else:
            names = self.ns
            ind = float(names.index(name))
            self.n = len(names)
            s = '%d/%d %.2f%%  %s time:%s'%(ind,self.n,
                                            (ind*100./self.n),name, times)
        if self.loged:
            log(s)
        return s
    log = __call__

class LogException():
    '''
    用于扑捉和记录函数中的异常错误
    usage:
        loge = LogException(logFilePath,otherFun) # 先实例一个对象
        loge.listen(f,*l,**args) # 用 .listen 运行函数
        @loge.decorator  # 装饰被监听函数
    '''
    def __init__(self,
                 logFilePath=None, 
                 otherFun=None,
                 printOnCmd=True, 
                 logBegin=False,
                 localTime=False,
                 isOn=True):
        '''logFilePath  log文件保存路径,为False时 不写入文件
        otherFun  一个返回字符串的函数，每次错误运行一次，结果写入log
        printOnCmd   时候在屏幕上打印
        logBegin     是否记录开始监听事件
        localTime    是否为当地时间 默认为GMT时间
        '''
        
        self.splitLine = '/*----------*/'
        self.path = logFilePath
        self.fun = otherFun if otherFun else lambda :''
        self.printt = printOnCmd
        self.localTime = localTime
        self.format = (
'''     Index :{index}, {time}
{timeClass} :{timeStr}
 Exception :{exceptionName}
   Message :{message}
      Args :{args}
{otherInfo}
{splitLine}
''')
        self.index = 0
        self.isOn = isOn
        if logFilePath:
            if os.path.isfile(logFilePath):
                with open(self.path,'r') as f:     
                    strr = f.read()
                self.index = strr.count(self.splitLine)
        if logBegin:
            class beginLogException(Exception):
                pass
            def f ():
                raise beginLogException('LogException is begin to log Exception!')
            self.listen(f)
        
    def listen(self,f,*l,**args):
        '''
        f        : 监听函数
        *l,**args: 函数f的参数
        '''
        import time
        if not self.isOn:
            f(*l,**args)
            return
        try:
            f(*l,**args)
        except Exception as e:
            exceptionName = tounicode(type(e))
            exceptionName = exceptionName[exceptionName.index('.')+1:-2]
            
            timeClass = 'Local time' if self.localTime else '  GMT time'
            t = time.asctime( time.localtime(time.time())) if self.localTime  else time.asctime(time.gmtime(time.time()))
            
            otherInfo = self.fun()
            
            erroStr = self.format.format(
                                         index=self.index,
                                         timeStr=t,
                                         timeClass=timeClass,
                                         time=time.time(),
                                         exceptionName=exceptionName,
                                         message=e.message,
                                         args=tounicode(e.args),
                                         otherInfo=otherInfo,
                                         splitLine=self.splitLine,
                                         )
            self.__writeLog(erroStr)
            self.index += 1
            self.last = self.e = e

    def __writeLog(self,strr):
        '''
        判断是否打印和写入
        '''
        if self.printt:
            print(strr)
        if self.path:
            with open(self.path,'a') as f:     
                f.write(strr)
    def decorator(self,f):
        '''
        函数装饰器封装
        '''
        def ff(*l,**arg):
            r = self.listen(f,*l,**arg)
            return r
        return ff

gsAttrDic = {}
class Gs(dict):
    '''debug的全局变量G的子类
    '''
    pass
    def __init__(self,name, log ,*l,**kv):
        idd = id(self)
        gsAttrDic[idd] = dicto()
        gsAttrDic[idd].log = log
        gsAttrDic[idd].name = name if name else 'g'
        dict.__init__(self,*l,**kv)
    def __setitem__(self, k, v):
        idd = id(self)
        if isinstance(v,tuple) and len(v) == 2 and (v[1] is self):
            v,tag = v
            if gsAttrDic[idd].log or (gsAttrDic[idd].log is None):
                pblue('%s[%s] ='%(gsAttrDic[idd].name, str(k)),v)
        else:
            if gsAttrDic[idd].log:
                pblue('%s[%s] ='%(gsAttrDic[idd].name, str(k)),v)
        dict.__setitem__(self, k, v)
        pass
    def logWhenSet(self, on='nothing'):
        idd = id(self)
        if on == 'nothing':
            gsAttrDic[idd].log = not gsAttrDic[idd].log
        else:
            gsAttrDic[idd].log = on
    def __call__(self, x=None):
        pretty = False
        if pretty:
            loc = prettyFrameLocation(1)
            pblue('Print by g from %s:'%loc)
        else:
            pblue('%s: %s'%(clf.p%'Print by g', x))
        return x
    __sub__ = __call__
    __lshift__ = __call__
    __rshift__ = __call__
    __truediv__ = __div__ = __call__
    
    def __del__(self):
        idd = id(self)
        if gsAttrDic and idd in gsAttrDic:
            del gsAttrDic[idd]
        
            
    

class SuperG(Gs):
    '''
    Author: DIYer22@GitHub (ylxx@live.com)
    **用于debug的全局变量**, 用了不少旁门左道以达到最低使用负担
    本身是类似JavaScript对象的超级字典，可以动态添加属性
    其中 g.keys,g.values,g.items 添加了property
    用法:
        1.获取变量
            g[0] = v 
            g.a = v
        2.在赋值时 同时打印
            g.a = v,g 
        3.始终在赋值时打印
            g.logWhenSet(True)
        4.替换print
            g(x)
            g<<x
            g-x
            g==x
        5.替换print 且返回本身
            g/x
    '''
    def __init__(self,name=None, log=None, *l, **kv):
        '''
        name: 变量名
        log: 时候打印 默认为None
            True: 都打印
            None: 跟着g才打印
            False: 都不打印
        '''
        Gs.__init__(self,name, log, *l, **kv)
    def __getattribute__(self, name=None, *l):
        if name in self:
            return self[name]
        if name in dir(Gs):
            return Gs.__getattribute__(self,name, *l)
        def keyError(x):
            self[name] = x
            return x
        return keyError
    def __setattr__(self, name, v):
        if name in dir(Gs):
            return Gs.__setattr__(self,name,v)
        self[name] = v
    if py2:
        @property
        def keys(self):
            return addCall(dict.keys(self))
        @property
        def values(self):
            return addCall(dict.values(self))
        @property
        def items(self):
            return addCall(dict.items(self))

sg = SuperG()


class withOperation():
    '''
    `w{operation}` is mulitple variables version of "{operation}/x", and work in "with statement".
    {usage}
    `w{operation}` only act on assignment variables that under `with w{operation}:` statement.
    
    Usage
    --------
        >>> with w{operation}:
        >>>     pi = 3.14
        >>>     e = 2.71
    
    Note
    --------
        If var's name in locals() and `id(var)` not change ,var may not be detected 
        Especially following cases：
            1. var is int and < 256
            2. `id(var)` not change

    '''
    def __init__(self, operation='p', printt=False, transport=False, deep=0):
        self.locs = defaultdict(lambda:[])
        self.operation = operation
        self.printt = printt
        self.transport = transport
        self.deep = deep
    def __enter__(self):
        f = sys._getframe(self.deep+1)
        ind = id(f)
        self.locs[ind].append((f.f_locals).copy())
        return self
    def __exit__(self, typee, value, traceback):
        f = sys._getframe(self.deep+1)
        ind = id(f)
        locsb = self.locs[ind].pop()
        locs = f.f_locals
        kvs = []
        newVars = []
        for k in locs:
            if k in locsb:
                if not (locsb[k] is locs[k]):
                    kvs.append((k, locs[k]))
            else:
                kvs.append((k, locs[k]))
                newVars.append(k)
        if self.transport:
            root = getRootFrame()
            root.f_locals.update(kvs)
        
        printf = lambda *l, **kv: 0
        if self.printt:
            printf = log
        printf("")
        printf(colorFormat.b%'withprint from %s'%prettyFrameLocation(f))

        if len(kvs):
            tag = ''
            if locs.get('__name__') == '__main__':
                printf(colorFormat.b% 'New Vars: ', end='')
                if len(newVars):
                    printf((', '.join([colorFormat.p%k for k in newVars])))
                else:
                    printf((colorFormat.b% 'None'))
                    
                tag = ("\nP.S. code run in __main__, some vars may not detected if id(var) not change.")
            printf((colorFormat.b% "All Vars's Values :"))#(P.S.some base object may not detected):"))
            if self.printt:
                from boxx import tree
                tree(dict(kvs))
                tag and printf(tag)
            
        else:
            print((colorFormat.r% '\n\nNot detected any Vars:'+
                   '\n    id(var) may not change in interactive mod if var is int and < 256 \n'+
                   '    `help(withprint)` for more infomation\n'+
                   '     P.S assignment self is not work for with statement.\n'+
                   '     Instead, `new_var = old_var` is OK!'))
    def __str__(self):
        s = self.__doc__
        if py2:
            return str(s.encode('utf-8'))
        return s
    __repr__ = __str__

class withPrint(withOperation):
    operation = 'p'
    usage = '''
    pretty print variables with their variable name.
    '''
    __doc__ = withOperation.__doc__.format(operation=operation, usage=usage)
    def __init__(self):
        withOperation.__init__(self, self.operation, True, False)
    
class withTransport(withOperation):
    operation = 'g'
    usage = '''
    `wg` will transport variable to Python interactive console.
    '''
    __doc__ = withOperation.__doc__.format(operation=operation, usage=usage)
    def __init__(self):
        withOperation.__init__(self, self.operation, False, True)

class withPrintAndTransport(withOperation):
    operation = 'gg'
    usage = '''
    `wgg` will transport variable to Python interactive console and pretty print they.
    '''
    __doc__ = withOperation.__doc__.format(operation=operation, usage=usage)
    def __init__(self):
        withOperation.__init__(self, self.operation, True, True)

wp = withPrint()
wg = withTransport()
wgg = withPrintAndTransport()

class TransportToRootFrame():
    def __init__(self, name=None, log=False):
        self.name = name
        self.log = log
    def __call__(self, value):
        frame = getRootFrame()
        frame.f_globals[self.name] = value
        if self.log:
            name = str(self.name)
            s = (clf.r%'gg.%s:"'+'%s'+clf.r%'"')%(name, tabstr(str(value), len(name)+5))
            print(s)
        return value
    __sub__ = __truediv__ = __div__ = __mul__ = __add__ = __eq__ = __pow__ = __call__
    def __str__(self):
        return 'TransportToRootFrame(name=%s, log=%s)'%(self.name, self.log)
    __repr__ = __str__
    
global_g_paras = {}
class GlobalGCore(object):
    def __init__(self, log=False):
        object.__init__(self)
        d = global_g_paras[id(self)] = dicto()
        d.log = log
        
        d.wo = withOperation(['g', 'gg'][log], printt=log, transport=True, deep=1)
    def __call__(self, deep=0):
        d = global_g_paras[id(self)]
        log = d.log
        out(depth=deep+1, printt=log)
        return True
        
    def __del__(self):
        idd = id(self)
        if global_g_paras and idd in global_g_paras:
            del global_g_paras[idd]

    def __enter__(self):
        d = global_g_paras[id(self)]
        return withOperation.__enter__(d.wo)
    def __exit__(self, typee, value, traceback):
        d = global_g_paras[id(self)]
        withOperation.__exit__(d.wo, typee, value, traceback)
class GlobalG(GlobalGCore):
    # TODO: add Decorators(装饰器) to catch error in with and export vars
    '''
    TODO:
        
    for dev-tips:
        after every operating in IPython@spyder, will read this instance 10+ times
        some times will read attr like "__xx__" but not in dir(object): in this case
        don't return anything just raise the Exception
        
        if is instance.__getattribute__(name) try to use getattr(instance, name) instead
    '''
    def __init__(self, log=False):
        GlobalGCore.__init__(self, log)
    def __getattribute__(self, name='x', *l):
#        print(id(self),name)
        if name.startswith('__') and name.endswith('__') or name in dir(GlobalGCore): 
            return GlobalGCore.__getattribute__(self, name, *l)
        log = global_g_paras[id(self)].log
        return TransportToRootFrame(name,log)
    def __setattr__(self, name, v):
        log = global_g_paras[id(self)].log
        transport = TransportToRootFrame(name,log)
        transport(v)
g = GlobalG()
gg = GlobalG(log=True)

config = dicto()
cf = config
boxxcf = dicto()

def if_main_then_g_call():
    '''`mg` is short of if_main_then_g_call
    Will transport `locals()` to Python interactive console when __name__ == '__main__'.
    Then return  __name__ == '__main__'
    
    >>> mg() 
    # equl to      
    >>> if __name__ == '__main__':
    >>>     g()
    '''
    frame = sys._getframe(1)
    _name_ = frame.f_globals.get('__name__')
    is_main = _name_ == '__main__'
    if is_main:
        g(1)
    return is_main
mg = if_main_then_g_call

def prettyClassFathers(obj):
    '''
    get object or type, return pretty str
    
    >>> prettyClassFathers(cf)
    Instance of boxx.tool.toolStructObj.dicto <-dict <-object
    '''
    fas = getfathers(obj)
    fas = [colorFormat.p%typeNameOf(fa) for fa in fas]
    s = 'Type' if istype(obj)  else 'Instance'
    s = colorFormat.r % s
    s += ' of '+ (' <-').join(fas)
    return s

def prettyFrameLocation(frame=0):
    '''
    get frame return pretty str
    
    >>> prettyFrameLocation(frame)
    "/home/dl/junk/printtAndRootFarme-2018.03.py", line 109, in wlf
    
    Parameters
    ----------
    frame : frame or int, default 0
        if int:相对于调用此函数frame的 int 深度的对应frame
    '''
    if isinstance(frame, int):
        frame = sys._getframe(1 + frame)
    c = frame.f_code
    return ((colorFormat.b%'File: "%s", line %s, in %s')%
            ('\x1b[32m%s\x1b[0m'% c.co_filename, '\x1b[32m%s\x1b[0m'% c.co_firstlineno, colorFormat.purple% c.co_name))

def getNameFromCodeObj(code, pretty=True):
    name = code.co_name
    filee = code.co_filename
    if pretty:
        if name == '<module>':
            if filee.startswith('<ipython-input-'):
                name = 'ipython-input'
            else:
                name = '%s'%os.path.basename(filee)
            name = '\x1b[36m%s\x1b[0m'%name
        if name == '<lambda>':
            return 'lambda'
    return name
def prettyFrameStack(frame=0, endByMain=True, maxprint=None):
    '''
    get frame return pretty str of stack
    
    >>> prettyFrameLocation(frame)
    __init__ <-f <-ff <-demo.py
    
    Parameters
    ----------
    frame : frame or int, default 0
        if int:相对于调用此函数frame的 int 深度的对应frame
    endByMain : bool, default True
        为 True 则在第一个 frame.f_locals[‘__name__’] == ‘__main__’ 处停止搜寻 
        目的是去除 IPython 自身多余的 Call Stack
    '''
    if frame is False and endByMain:
        frame, endByMain = 0, False
    if isinstance(frame, int):
        frame = sys._getframe(1 + frame)
    fs = getFatherFrames(frame, endByMain=endByMain)
    ns = [getNameFromCodeObj(f.f_code) for f in fs]
    if endByMain:
        if 'execfile' in ns:
            ns = ns[:ns.index('execfile')]
        if '_call_with_frames_removed' in ns:
            ns = ns[:ns.index('_call_with_frames_removed')]
    s = ' <-'.join(ns)
    s = shortStr(s, maxlen=maxprint)
    return s

def generaPAndLc():
    saveOut = {}
    class LocalAndGlobal(dicto):
        '''
        your can use `out()`, `p()` in any function or module
        exec `out()`, all vars that belong the function will transport to
        Python interactive shell. and `globals()` will in `p` which is a dicto
        
        BTW, `import boxx.out`, `import boxx.p` is Convenient way to use `out()` without `from boxx import out`
        
        在函数内运行`p()` or `lc()`  
        则此函数的global和local 变量会载入全局变量 p 中
        函数的 frame等其他信息 则放入全局变量 lc
        
        Parameters
        ----------
        depth : int or bool, default 0
            相对于`p()`所在frame的深度
            ps.为了使关闭printt简便 若`p(False)` 自动转换为`p(0, False)`
        printt : bool, default True
            是否打印
            
        Effect
        ----------
        out() : dicto
            copy current frame's locals() and globals() to your 
            Python interactive shell
            That's mean any var befor `out()` will Transport to Python shell,
            Even your are exec `out()` in Thearding
            ps. `out(False)` will turn off the info print
        p() : dicto
            将当前frame的locals()和globals()存入p
            p.var_name is the var's value in locals() and globals()
        lc : callabel dicto
            lc则存储更多、更细致的信息 包含code, frame, frames栈
            lc 的 items：
                self.c = self.code = frame.f_code
                self.l = self.local = locals()
                self.f = self.frame = frame
                self.fs = getFatherFrames(frame) 
        '''
        def __init__(self, out=False):
            saveOut[id(self)] = out
        def __call__(self, depth=0, printt=True):
            if depth is False:
                printt = depth
                depth = 0
            frame = sys._getframe(depth+1)
            code = frame.f_code
            p.clear()
            glob = frame.f_globals
            local = frame.f_locals
                
            self.c = self.code = code
            self.f = self.frame = frame
            self.l = self.local = local
            if printt:
                print('')
                prettyStr = prettyFrameLocation(frame)
                print(prettyStr)
                fs = getFatherFrames(frame)
                self.fs = fs
                s = prettyFrameStack(frame)
                print((colorFormat.b%'Stacks: '+colorFormat.r%s))
                print((colorFormat.b%'Locals: '))
                from boxx import tree
                tree(local, maxprint=100)
                
            if saveOut[id(self)]:
                p.update(glob)
                p.update(local)
                root = getRootFrame()
                addDic = dict(
#                        code=code,
#                        frame=frame,
#                        local=local,
#                        glob=glob,
#                        gl=glob,
#                        rootGlob=root.f_globals,
                        )
                
                
                addDic.update(local)
                same = set(addDic).intersection(set(root.f_globals))
                
                if printt:
                    print("")
                    addVarStr = ', '.join([colorFormat.p%k for k in addDic if k not in same])
                    if addVarStr:
                        print(colorFormat.b% '\nVars add to Console Frame: '+'\n└── '+ addVarStr)
                    if len(same):
                        print(colorFormat.r% '\nVars that replaced in Console Frame: '+'\n└── '+', '.join([colorFormat.p%k for k in same]))
                root.f_globals.update(addDic)
                lc.c = lc.code = code
                lc.f = lc.frame = frame
                lc.l = lc.local = local
                self.clear()
                self.update(local)
    
    
    __P_CACHE__ = dicto()
    class Pdicto(dicto):
        '''
        # print(x) and return x
        >>> p/517 
        517
        517
        
        # p() to pretty print all variables in function with thier name
        >>> def f(arg=517):
                l = [1, 2]
                p()
        
            ├── l: list  2
            │   ├── 0: 1
            │   └── 1: 2
            └── arg: 517
        import boxx.p has same usage
        
        # p will pretty print mulit variables under "with statement"
        
        >>> with p:
            
        '''
        def __call__(self, depth=0, printt=True):
            if depth is False:
                printt = depth
                depth = 0
            lc(depth+1, printt)
        def printt(self, x=None):
            pretty = False
#            root = getRootFrame()
#            root.f_globals['pp'] = x
            if pretty:
                loc = prettyFrameLocation(1)
                pblue('Print by p from %s:'%loc)
            else:
                pass
#                pblue('%s: %s'%(clf.p%'As pp by p', x))
                print(x)
            return x
        __sub__ = printt
        __lshift__ = printt
        __rshift__ = printt
        __truediv__ = __div__ = printt
        __pow__ = printt
        
        def __enter__(self):
            d = __P_CACHE__
            if 'wo' not in d:
                d.wo = withOperation('p', printt=True, transport=False, deep=1)
            return withOperation.__enter__(d.wo)
        def __exit__(self, typee, value, traceback):
            d = __P_CACHE__
            withOperation.__exit__(d.wo, typee, value, traceback)
    p = Pdicto()
    lc = LocalAndGlobal()
    out = LocalAndGlobal(out = True)
    return p, lc, out
p, lc, out = generaPAndLc()
pp = "registered `pp` var name that will be used by `p/x`"
#pp, lcc, outt = generaPAndLc()




# -*- coding: utf-8 -*-

import os,glob
from ..ylsys import tmpboxx

def showImgsInBrowser(paths, htmlp=None, coln=None):
    '''图片展示分析工具  使用浏览器同步显示图片的image,gt,resoult 等
    支持放大缩小与拖拽
    
    Parameters
    ----------
    paths : list of path
        图片地址组成的数组
    '''
#    paths = list(map(os.path.abspath,paths))
    html = getShowsHtml()
    s = 'srcs = [%s]'%(",".join(['"%s"'%p for p in paths]))
    if coln is None:
        # if len(paths) in (0, 1, 2, 4):
        coln = 1
        if len(paths) >= 2:
            coln = 2
        if len(paths) == 3 or len(paths) > 4:
            coln = 3
        if len(paths) >= 12:
            coln = 4
    s += "; coln = %s" % int(coln)
    html = html.replace('//replaceTagForPython',s.replace('\\',r'\\'))
    if not htmlp:
        htmlp = os.path.join(tmpboxx(), 'shows-%03d.html') %len(glob.glob(os.path.join(tmpboxx(), 'shows-*.html')))
    with open(htmlp,'w') as f:
        f.write(html)
    from boxx import browserOpen
    browserOpen(htmlp)

def getShowsHtml():
    # TODO: 
    """
    - middle mouse button move left right to rescale and +/- to rescale 
    - support scale by pixel and scale by max(hw) for different image size
    - independent as .html and support ?imgs=[]&coln=3
    - gif mode with time gap=1
    - download/copy img jpg (if gif mode, copy gif)
    - support touch screen to move and scale
    """
    
    return '''
    <!DOCTYPE html>
    <html>
    <head>
      <meta charset="UTF-8">
      <title>boxx.shows(images)</title>
      <script src="https://cdn.bootcss.com/dat-gui/0.6.5/dat.gui.js"></script>
      <style>
      </style>
      <script>
      log = console.log
      int = parseInt
      float = parseFloat
      pow = Math.pow
      max=(x)=>{
        var m=x[0]
        for (var i = 0; i < x.length; i++) {
            m=x[i]>m?x[i]:m
        };
        return m
      }
      min=(l)=>-max(l.map((x)=>-x))
      len= (x)=>x.length
      </script>
    </head>
    <body style="margin:0;padding:0; overflow:hidden;">
    	<canvas id="canvas" width="" height="" style="top:0px; left:0px;">
    	</canvas>
    </body>
    <script>
    canvas = document.getElementById('canvas')
    ctx = canvas.getContext('2d');

    window.addEventListener("resize", resizeCanvas, false);

    function resizeCanvas() {
        canvas.width = window.innerWidth;
        var n = len(srcs)
        bys = int(n/coln)+Boolean(n%coln)
        if(bys<=coln){
          canvas.height = window.innerHeight;
        }else(
          canvas.height = window.innerWidth*bys/coln
        )
        var w=canvas.width
        var h=canvas.height

        sw = w/coln
        sh = h/bys
        flash();
    }
    busy = 0
    roundTag = 0
    flashInner = ()=>{
      roundTag += 1
      if (busy){
        return
      }
      ctx.imageSmoothingEnabled = isSmooth
      var tag = roundTag
      busy = 1
      var bx=0
      var by=0
      for (var src_idx of Array(coln*bys).keys()) {
        if (roundTag!=tag){
          busy=0;
          flash()
          return
        }
        if(isRoll){
          old = src_idx
          sourceBx = (bx-rollx)%coln
          if(sourceBx<0){sourceBx += coln}
          sourceBy = (by-rolly)%bys
          if(sourceBy<0){sourceBy += bys}
          src_idx = sourceBy * coln + sourceBx
                }
        if (src_idx < srcs.length){
                var src = srcs[src_idx]
                drawImage(src,bx,by)
        }else{
            ctx.clearRect(bx*sw,by*sh,sw,sh);
            }
        bx += 1
        if(bx==coln){
          by+=1
          bx=0
        }
      }
      busy=0;
      if (roundTag!=tag){
        flash()
      }
    }
    flash = ()=>{setTimeout(flashInner,0)}

    drawImage = (src,bx,by)=>{
      var img = new Image();
      img.onload = function(){
        imgh = img.height
        imgw = img.width
        if(scale == -1){
          sca = min([sw/imgw,sh/imgh])
          scale = Math.log2(sca)
        }
        sca = pow(2,scale)
        var left = max([0,imgw*fx-sw/sca/2])
        var right = min([imgw,imgw*fx+sw/sca/2])
        var up = max([0,imgh*fy-sh/sca/2])
        var down = min([imgh,imgh*fy+sh/sca/2])
        var leftt = max([0,sw/2-imgw*fx*sca])
        var rightt = min([sw,sw/2+imgw*(1-fx)*sca])
        var upp = max([0,sh/2-imgh*fy*sca])
        var downn = min([sh,sh/2+imgh*(1-fy)*sca])
        ctx.clearRect(bx*sw,by*sh,sw,sh);
        ctx.drawImage(img,left,up,right-left,down-up,   bx*sw+leftt,by*sh+upp,rightt-leftt,downn-upp);
          // ctx.drawImage(img,0,0,6000,6000,   0,0,1000,1000);
      }
      img.src = src;
    }

    var view = {
        scroll:(e)=>{
            e.preventDefault()
            var x=e.offsetX
            var y=e.offsetY
            var change = int(e.deltaY)
            var zoomSpeed = 0.2
            if (change<0){
                scale += (zoomSpeed)
            }
            if (change>=0){
                scale -= (zoomSpeed)
            }
            flash()
        },
        onMove:(e)=>{
            e.preventDefault()
            var x = e.clientX-xBegin
            var y = e.clientY-yBegin
            var speed = 2
            fx = fxBegin - x/sca/imgw*speed
            fx = min([1,max([0,fx])])
            fy = fyBegin - y/sca/imgh*speed
            fy = min([1,max([0,fy])])
            flash()
        },

        onDown:(e)=>{
            e.preventDefault()
            var x = e.clientX
            var y = e.clientY
            xBegin = x
            yBegin = y
            fxBegin = fx
            fyBegin = fy
            canvas.onmousemove = view.onMove
        },
        onUp:()=>{
            canvas.onmousemove = null
        },
        mouseBegin:()=>{
            window.onmousewheel=view.scroll
            window.onmousedown = view.onDown
            window.onmouseup = view.onUp
            onmouseup()
        }
    }
    scale = -1
    fx = .5
    fy = .5
    rollx = -1
    rolly = 0
    isRoll = false
    isSmooth = false

    var srcs = []
    coln = 2
    //replaceTagForPython
    
    resizeCanvas()
    view.mouseBegin()
    var gui = new dat.GUI();
    var con = gui.add(window, 'scale',-10,5).listen().name('scale 🖱wheel');
    con.onChange( (v)=>{
        flash()
    })

    // var con2 = gui.add(window, 'fx',0.,1.0).listen();
    // var con3 = gui.add(window, 'fy',0.,1.0).listen();
    // con2.onFinishChange = con3.onFinishChange = con.onFinishChange

    var con4 = gui.add(window, 'coln',1,5).step(1).listen();
    con4.onChange ( function(v){
      coln = int(v)
      resizeCanvas()
    }).name('colN ⌨️123')
    
    gui.add(window,'isRoll').listen().onChange(()=>{
        flash()
        }).name('isRoll ⌨space/Z')
    gui.add(window,'rollx', -srcs.length, srcs.length).step(1).listen().onChange(()=>{
        flash()
        }).name('rollX ⌨️AD')
    gui.add(window,'rolly', -srcs.length, srcs.length).step(1).listen().onChange(()=>{
        flash()
        }).name('rollY ⌨️WS')
    
    gui.add(window,'isSmooth').listen().onChange(()=>{
        flash()
        })
    
    HOLDING_KEYS = {}
    document.onkeydown=(e)=>{
        var e = e || window.event;
        //console.log(e)
        if (e.ctrlKey || e.shiftKey || e.altKey || e.metaKey){
                return
                }
        if(e.key == " "){
                isRoll = !isRoll
                flash()
                }
        if("zZ".includes(e.key)){
                if (!HOLDING_KEYS["z"]){
                    HOLDING_KEYS["z"]=true
                    isRoll = !isRoll
                    flash()
                    }
                }
        if("aA".includes(e.key)){
                rollx -= 1
                flash()
                }
        if("dD".includes(e.key)){
                rollx += 1
                flash()
                }
        if("wW".includes(e.key)){
                rolly += 1
                flash()
                }
        if("sS".includes(e.key)){
                rolly -= 1
                flash()
                }
        if("1".includes(e.key)){
                coln = 1
                resizeCanvas()
                }
        if("2".includes(e.key)){
                coln = 2
                resizeCanvas()
                }
        if("3".includes(e.key)){
                coln = 3
                resizeCanvas()
                }
        if("4".includes(e.key)){
                coln = 4
                resizeCanvas()
                }
        if("5".includes(e.key)){
                coln = 5
                resizeCanvas()
                }
        if("6".includes(e.key)){
                coln = 6
                resizeCanvas()
                }
        }
    
    document.onkeyup=(e)=>{
        var e = e || window.event;
        if("zZ".includes(e.key)){
                isRoll = !isRoll
                flash()
                HOLDING_KEYS["z"]=false
                }
        }
    </script>
    </html>
    '''
if __name__ == '__main__':
    # test code
    # img=imread('../../test/imgForTest/Lenna.jpg');shows([img, img.mean(-1), img//3, r])
    pass



# -*- coding: utf-8 -*-

#import sys
#from os.path import abspath,join,dirname
#_path = (abspath(__file__))
#absLibpPath = join(dirname(_path),'yl')
#    
#if absLibpPath not in sys.path:
#    sys.path = [absLibpPath]+sys.path

# print absLibpPath
from boxx.tool import *
from boxx.ylimg import *
from boxx.ylml import * 
from boxx.ylsci import *
from boxx import tool
from boxx import ylimg
from boxx import ylml
from boxx import ylnp


if __name__ == '__main__':

    pass



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import os
from ..ylcompat import interactivePlot
from ..tool.toolLog import LogLoopTime,log
from ..tool import pathjoin, getFunName
from ..ylsci import savenp, loadnp
import pandas as pd
import numpy as np


def accEvalu(re, gt):
    '''评估二分类分割问题 返回dict 包含 acc 和 loss
    
    Parameters
    ----------
    re : np.ndarray
        resoult or prob, 标签或h*w*n 的概率矩阵
    gt : np.ndarray of bool
        ground Truth, 值为每个像素的类别
    '''
    if re.ndim == 3:
        re = re.argmax(2)
    acc = (re==gt).sum()/float(gt.shape[0]*gt.shape[1])*100
    return {'acc':acc,'loss':100-acc}
    

def lplrEvalu(re, gt):
    '''评估二分类分割问题 返回dict 包含 ['LP', 'LR', 'OP', 'OR','me']'''
    re = re > 0
    TPl = (re==0)*(gt==0) #(re+gt) == 0
    FPl = (re==0)*(gt==1)
    FNl = (re==1)*(gt==0)
    

    TPs = (re==1)*(gt==1) 
    FPs = (re==1)*(gt==0)
    FNs = (re==0)*(gt==1)
#    show([TPs,FPs,FNs])
    TPl,FPl,FNl,TPs,FPs,FNs = [float(i.sum()) for i in [TPl,FPl,FNl,TPs,FPs,FNs]]
    LP = TPl/(TPl+FPl)
    LR = TPl/(TPl+FNl)
    OP = (TPl+TPs)/(TPl+FPl+TPs+FPs)
    OR = (TPl+TPs)/(TPl+FNl+TPs+FNs)
#    return (LP*100,LR*100,OP*100,OR*100)
    mean = np.mean([LP,LR,OP,OR])
    return {k:100*(1-v) for k,v in zip(['LP', 'LR', 'OP', 'OR','me'],[LP,LR,OP,OR,mean])}

def diceEvalu(re,gt):
    ''' Dice coefficient 方法评估二分类分割问题 返回dict 包含 ['dice', 'loss']
    
    Parameters
    ----------
    re : np.ndarray
        resoult or prob, 标签或h*w*n 的概率矩阵
    gt : np.ndarray of bool
        ground Truth, 值为每个像素的类别
        '''
    if re.ndim == 3:
        re = re.argmax(2)
    dice = (re*gt).sum()*2/float(re.sum()+gt.sum())*100
    return {'dice':dice,'loss':100-dice,}



class Evalu(pd.DataFrame):
    '''
# 用于验证的工具类Evalu
继承自pandas.DataFrame， 会自动保存每个样本 resoult 及其评估结果，能对评估结果进行分析
并能以DataFrame的形似对评估结果操作与分析
常用缩写：
    * evalu = Evalu的实例
    * re = resoult, 需要评测的样本 
    * gt = GroundTruth, 用于评测的真值 
    * key = 评测项的名称 即列的名称 属于 self.columns
    * df = pd.DataFrame 实例
    * prob = 分割的概率矩阵 ，即shape为 h*w*n 的矩阵(n 表示类别数)
    
功能
--------
自动log :
    对每个样本 自动log出当前进度，当前迭代花费的时间，及评估结果
保存resoult :
    将用来评估的resoult保存为压缩的npz格式
    通过 evalu.re(name) 或者evalu[name] 来调用
载入之前的评估结果 :
    将loadcsv=True 或者 对应评估结果csv的路径即可载入
分析与可视化 :
    总结 :evalu.summary()
    可视化评估项的分布 :evalu.distr(key)
一键删除验证产生的所有文件 :
    evalu.removeFiles()
继承 pandas.DataFrame 的所有操作 :
    以df形似 分析evalu后的结果 
    ps:pandas快速入门 http://wiki.jikexueyuan.com/project/start-learning-python/311.html

Examples
--------
>>> names = glob("/home/[to val path]/*.jpg")
e = Evalu(binaryEvalu,
          evaluName='try-example',
          valNames=names,
          logFormat='acc:{acc:.4f}, loss:{loss:.4f}',
          sortkey='acc'
          )
for name in names:
    [get re,gt code]
    e.evalu(re,gt,name)
e.summary()
e.distr()
    '''
    def __init__(self,evaluFun,
                 evaluName='null',
                 valNames=None,
                 logFormat=None,
                 sortkey=None,
                 loadcsv=False,
                 saveResoult = False,
                 loged=True,
                 savepath='./val/',
                 ):
        '''
Parameters
----------
evaluFun : function
    用于评估resoult的函数，此函数须接受两个参数 (re,gt),并返回一个dict
    dict 包含每一个评估项的值
    例如 evaluFun(re,gt) => return {'key1':value1,'key2':value2}
evaluName : str, default 'null'
    实例的名称
valNames : list or tuple, default None
    验证集所有样本的名字
logFormat : str, default None
    fromat格式 用于规范评估结果的显示效果  
    会执行 print logFormat.format(**evaluFun(re,gt))
    如: 'key1 is {key1:.2f}, key2 is {key2:.2f},'
    默认为直接打印 evaluFun(re,gt)
sortkey : str, default None
    用于排序时候的key 默认为df.columns[-1]
loadcsv : bool or str, default False
    载入已保存的csv 默认不载入，为True时候则载入，
    为str时候则载入str对应的path的csv
saveResoult : bool or function, default False
    是否保存resoult 默认为False ,True  则保存 re
    若为function 则保存 saveResoult(re) 为 .npz格式
loged : bool, default True
    是否每次评估都打印出结果 默认为打印
savepath : str, default './val/'
    保存的路径 若loadcsv 为path 则为 dirname(loadcsv)

Examples
--------
e = Evalu(accEvalu,
          evaluName='binary segment evalu',
          valNames=names,
          logFormat='acc:{acc:.4f}, loss:{loss:.4f}',
          sortkey='loss'
          )
        '''
        pd.DataFrame.__init__(self,)
        self._inited = False
        self.evaluName = evaluName
        self.evaluFun = evaluFun
        self.logFormat = logFormat
        self.logLoop = LogLoopTime(valNames,loged=False)
        self.n = -1 if valNames is None else len(valNames)
        self.sortkey = sortkey
        self.loged = loged
        self.saveResoult = saveResoult
        self.__isTuple = False
        
        if isinstance(loadcsv,str):
            self.evaluDir = os.path.dirname(loadcsv)
        else:
            self.evaluDir = os.path.join(savepath,'Evalu-%s'%evaluName)
        self._defaultCsvPath = pathjoin(self.evaluDir,self.evaluName+'.csv')
        if loadcsv :
            if isinstance(loadcsv,str):
                self._defaultCsvPath = loadcsv 
            self._loaddf()
    def _initdf(self,name,dic):
        self._inited = True
        dic['name'] = name
        pd.DataFrame.__init__(self,[pd.Series(dic)])
        self.set_index('name',inplace=True)
        if self.sortkey is None:
            self.sortkey = self.columns[-1]
    def _loaddf(self):
        self._inited = True
        pd.DataFrame.__init__(self,pd.read_csv(self._defaultCsvPath,index_col='name'))
        if self.sortkey is None:
            self.sortkey = self.columns[-1]
        
    def __makedirs(self):
        path = pathjoin(self.evaluDir,'npzs')
        if not os.path.isdir(path):
            os.makedirs(path, exist_ok=True)
    def _log(self,s):
        if self.loged:
            log('%s'%s)
    def evalu(self, re, gt, name=None):
        '''使用 self.evaluFun 评估 resoult GrountdTruth
        
        Parameters
        ----------
        re : as parameters of self.evaluFun
            需要评测的resoult
        gt : as parameters of self.evaluFun
            用于评测的真值 GroundTruth
        name : str, default None
            记录此次评估样本的名称 为空则不记录    
        '''
        dic = self.evaluFun(re,gt)
        s = self._formatDic(dic)
        if name is None:
            self._log(s)
            return dic
        if not self._inited:
            self._initdf(name,dic)
        else:
            self.loc[name] = dic
        if self.logLoop:
            loopLog = self.logLoop(name)
            s = (loopLog+' Evalu:'+s)
        self._log(s)
        if len(self)==self.n and self.index[-1] == name and (
                self.loged or self.saveResoult):
            self.savecsv()
        if self.saveResoult:
            self._savere(name,re)
        return dic
    def savecsv(self,):
        self.__makedirs()
        self.to_csv(self._defaultCsvPath)
    def _savere(self,name,re):
        self.__makedirs()
        npname = pathjoin(self.evaluDir,'npzs',name)
        if '__call__' in dir(self.saveResoult) :
            savenp(npname,self.saveResoult(re))
        else:
            if re.ndim == 3:
                re = np.uint8(re.argmax(2))
            savenp(npname,re)
    def re(self,name=-1):
        '''return np.array, 返回自动保存下来的np.array 样本
        
        Parameters
        ----------
        name : str or int, default -1
            当为int 时 name=df.index[int]    
        '''
        if isinstance(name,int):
            name = self.index[name]
        npname = pathjoin(self.evaluDir,'npzs',name)
        re = loadnp(npname)
        return re
    def __getitem__(self,name):
        '''同 Evalu.re
        return np.array, 返回自动保存下来的np.array 样本
        
        Parameters
        ----------
        name : str or int, default -1
            当为int 时 name=self.index[int]  
        '''
        if name in self.columns:
            return pd.DataFrame.__getitem__(self,name)
        return  self.re(name)
    
    def _sorted(self,key,ascending=True):
        if isinstance(key,int) and key not in self.columns:
            key = self.columns[key]
        if key is None:
            key = self.sortkey
        return self.sort_values(key,ascending=ascending)
    @property
    def highh(self):
        return self.high()
    def high(self,n=10,key=None):
        '''return df, 评估项key 最高的n个样本组成的df
        
        Parameters
        ----------
        n : int, default 10
            返回的样本数目   
        key: key in self.columns, default self.sortkey
            用于计算排名的 key
        '''
        df = self._sorted(key,ascending=False)
        return df.head(n)
    @property
    def loww(self):
        return self.low()
    def low(self,n=10,key=None):
        '''与evalu.high 相反 参见evalu.high
        '''
        df = self._sorted(key)
        return df.head(n)
    @interactivePlot
    def distr(self, key=None):
        '''分析key 的分布，describe + 分布图
        
        Parameters
        ----------
        key : str or int, default self.sortkey
            当为int 时 key=df.columns[int]
        '''
        import matplotlib.pyplot as plt
        if isinstance(key,int) and key not in self.columns:
            key = self.columns[key]
        key = self.sortkey if key is None else key 
        df = self._sorted(key,False)
        print((df[key].describe()))
        df['_index_'] = list(range(1,len(df)+1)) 
        df['__mean'] = mean = df[key].mean()
        df.set_index('_index_',inplace=True)
        df['__mean'].plot(style='orange')
        df[key].plot.bar(title='mean = %s'%mean)
        plt.show()
    
    def _formatDic(self,dic):
        formatt = self.logFormat
        if formatt:
            s=(formatt.format(**dic))
        else:
            s=(', '.join(['%s: %s'%(k,v) for k,v in list(dic.items())]))
        return s
    def __topAndLowStr(self,n=10):
        top = self.high(n)
        low = self.low(n)
        tops = str(top[self.sortkey]).split('\n')
        title , tops = tops[-1], tops[1:-1]
        lows = str(low[self.sortkey]).split('\n')[1:-1]
        strn = len(lows[0])+1
        second = ('high %d'%n+strn*' ')[:strn]+'| '+('low %d'%n)
        body='\n'.join([t+' | '+l for t,l in zip(tops,lows)])
        s= '''evaluName: %s, sortkey%s
%s
%s
'''%(self.evaluName, title[4:], second, body)
        return (s)
    def __str__(self,n=3):
        funName = getFunName(self.evaluFun)
        describe = self.describe()
        s = '''Evalu name       : %s
cache resoult dir: %s
sort key         : %s
evalu function  : %s
logFormat        : %s
        ''' %(self.evaluName,self.evaluDir,self.sortkey,funName,self.logFormat)
        
        return s+'\n'+self.__topAndLowStr(n)+str(describe)
    __repr__ = pd.DataFrame.__str__
    def summary(self,n=10):
        '''various summary 
        n:同时显示high和low的数量'''
        print(self.__str__(n))
        
    class Series(pd.Series):
        '''改造了 __str__ 的 pd.Series'''
        def __init__(self,series,strr):
            pd.Series.__init__(self,series)
            self.strr = strr
        def __str__(self):
            return self.strr
        __repr__ = __str__
    def __call__(self,name=-1,sortkey=None):
        '''return pd.Series, 对应name的评估结果
        
        Parameters
        ----------
        name : str or int, default -1
            当为int 时 name=df.index[int]    
        sortkey: key in self.columns, default self.sortkey
            用于计算排名的 key
        '''
        if isinstance(name,int):
            name = self.index[name]
        key = sortkey or self.sortkey
        row = self.loc[name]
        small,n = (self[key]<=row[key]).sum(),float(len(self))
#        row['high_in_[%s]'%key] = '%d/%d (%.2f%%)'%(small,n,small/n*100)
        dic = dict(row)
        s = "Name: %s  High in ['%s']:"%(name,key) + '%d/%d(%.2f%%)  Evalu:'%(small,n,small/n*100)
        s += self._formatDic(dic)
        return Evalu.Series(row,s)
    @property
    def df(self):
        return pd.DataFrame(self)
    
    def removeFiles(self):
        '''删除使用此Evalu产生的文件 即文件夹 self.evaluDir'''
        import shutil
        shutil.rmtree(self.evaluDir)
if __name__ == '__main__':
    e = Evalu(int)
    import pandas as pd
    df = pd.DataFrame({
                       0:list(range(5)),
                       1:list(range(10,15)),
                       'a':list("abcde"),
                       })
    df.set_index(0,inplace=True)


    pass



# -*- coding: utf-8 -*-

from __future__ import unicode_literals, print_function

import time
import inspect

from ..ylsys import cpun, winYl
from .toolStructObj import typestr

from functools import reduce

def getFunName(fun):
    '''
    获得函数的名称
    '''
    if '__name__' in dir(fun):
        return fun.__name__
    return '[unkonw]'

class SaveArguments:
    """
    Save Arguments in funcation and apply those arguments to fun
    """
    def __init__(self, *l, **kv):
        self.l = l
        self.kv = kv

    def apply(self, func):
        """
        Apply the saved arguments to the function
        """
        return func(*self.l, **self.kv)
        
    __call__ = apply
    def __str__(self):
        ls = ", ".join(map(str, self.l))
        if len(self.l) and len(self.kv):
            ls += ", "
        kvs = ", ".join(["%s=%s" % (kv) for kv in self.kv.items()])
        return "SaveArguments(%s)" % (ls + kvs)

    __repr__ = __str__



_NONE_TAG = []

class FuncArgs:
    """
    Catch and manipulate function argments, for example::
    
        def func(a=None, b=None, c=None, *args, **kwargs):
            print(a, b, c, args, kwargs)
        fa = FuncArgs(func, "a1", b="b1", c=3.14,)
        fa.apply(func)
        fa.set_arg(0, 0)
        fa.set_arg("b", "bb")
        fa.set_arg(float, 2.71)
        fa.apply(func)
    """

    class DummyBoundArguments:
        def __init__(self, args, kwargs):
            self.args = args
            self.kwargs = kwargs

        def find_argname(self, arg):
            if isinstance(arg, (tuple, type)):
                for idx, v in enumerate(self.args):
                    if isinstance(v, arg):
                        return idx
                for k, v in self.kwargs.items():
                    if isinstance(v, arg):
                        return k
            return arg

        def op_arg(self, key, value=_NONE_TAG):
            argname = self.find_argname(key)
            if isinstance(argname, int):
                if value is _NONE_TAG:
                    return self.args[argname]
                else:
                    self.args = (
                        self.args[:argname] + (value,) + self.args[argname + 1 :]
                    )
                    return
            if isinstance(argname, str):
                if value is _NONE_TAG:
                    return self.kwargs[argname]
                else:
                    self.kwargs[argname] = value
                    return

        def __str__(self):
            return "<DummyBoundArguments (args=%s, kwargs=%s)>" % (
                self.args,
                self.kwargs,
            )

    def __init__(self, func, *args, **kwargs):
        if not callable(func):

            def func(*args, **kwargs):
                pass

        self.func = func
        try:
            self.sig = inspect.signature(func)
            # inspect.BoundArguments
            self.ba = self.sig.bind(*args, **kwargs)
            self.ba.apply_defaults()
        except ValueError:
            self.ba = self.DummyBoundArguments(args, kwargs)

    def find_argname(self, arg):
        if not hasattr(self, "sig"):
            return self.ba.find_argname(arg)

        if isinstance(arg, str):
            argname = arg
        elif isinstance(arg, int):
            argname = list(self.sig.parameters.keys())[arg]
        elif isinstance(arg, (tuple, type)):
            # return first argname belong to arg type
            type_names = []  # 
            for idx, _arg in enumerate(self.ba.args):
                type_names.append(getattr(type(_arg), "__name__", "#"))
                if isinstance(_arg, arg):
                    argname = list(self.sig.parameters.keys())[idx]
                    return argname
            for k, _arg in self.ba.kwargs.items():
                type_names.append(getattr(type(_arg), "__name__", "#"))
                if isinstance(_arg, arg):
                    return k
            # __import__("IPython").embed()
            raise AssertionError("None of args's type%s belong to target type %s" % (type_names, str(arg), ))
        return argname

    def set_arg(self, key, value):
        if not hasattr(self, "sig"):
            self.ba.op_arg(key, value)
            return self
        argname = self.find_argname(key)
        self.ba.arguments[argname] = value
        return self

    def get_arg(self, key):
        if not hasattr(self, "sig"):
            return self.ba.op_arg(key)
        argname = self.find_argname(key)
        return self.ba.arguments[argname]

    def apply(self, func=None):
        if func is None:
            func = self.func
        return func(*self.ba.args, **self.ba.kwargs)

    __call__ = apply

    def __str__(self):
        s = str(self.ba)
        s = s.replace("BoundArguments ", "FuncArgs of " + self.func.__name__)
        return s

    __repr__ = __str__
    @staticmethod
    def test():
        def func(a=None, b=None, c=None, *args, **kwargs):
            print(a, b, c, args, kwargs)

        fa = FuncArgs(func, "a1", b="b1", c=3.14,)
        fa.apply(func)
        fa.set_arg(0, 0)
        fa.set_arg("b", "bb")
        fa.set_arg(float, 2.71)
        fa.apply(func)

def dynamicWraps(func):
    '''
    decorator 动态规划 装饰器
    '''
    cache={}
    from functools import wraps
    @wraps(func)
    def wrap(*args,**kv):
        key = args+tuple(sorted(kv.items())) if len(kv) else args
        if key not in cache:
            cache[key]=func(*args, **kv)
        return cache[key]
    return wrap

def pipe(*funList):
    '''
    将一组函数组成一个管道函数
    
    Parameters
    ----------
    funList : Function or list of Function
        函数list
    '''
    if len(funList) == 1 and  '__iter__' in dir(funList[0]):
        funList = funList[0]
    return reduce(lambda f1,f2:(lambda x:f2(f1(x))), funList, lambda x:x)
    
def setTimeout(fun, t=0):
    '''
    same to setTimeout in JavaScript
    '''
    from threading import Timer
    thread = Timer(t,fun)
    thread.start()
    return thread

def setInterval(fun, inter, maxTimes=None):
    '''
    same to setInterval in JavaScript
    '''
    maxTimes = [maxTimes]
    def interFun(): 
        fun()
        if maxTimes[0] is not None:
            maxTimes[0] -= 1
            if maxTimes[0] <= 0:
                return 
        setTimeout(interFun, inter)
    interFun()

from multiprocessing import Pool as PoolMp
from multiprocessing.dummy import Pool as PoolThread
def __multiprocessingFun__(args):
    '''
    多线程专用
    '''
    return args[0](*args[1:])

def __multiprocessLogFun__(args):
    '''
    '''
    
    fun, args, ind, lenn, logf = args
    if ind is None:
        return fun(*args)
    from .toolLog import percentStr, shortStr
    from .toolSystem import timeit
    with timeit(None) as t:
        re = fun(*args)
    if logf:
        logf([fun, args, ind, lenn, logf])
    else:
        print('%s/%s(%s) time: %s, args[0]: %s'%(ind,lenn,isinstance(lenn, int) and percentStr(ind*1./lenn),t.s,shortStr(args[0])), end='\n')
    return re

def mapmp(fun, *iterables, **kv):
    '''
    Map with Multi Processing
    mapmp(fun, sequence[, sequence, ...], pool=None, thread=False)->list
    
    !Important, multi processing must do in `__name__ == '__main__'`'s block 
    see more at https://docs.python.org/3/library/multiprocessing.html
    
    >>> mapmp(np.add, range(3), range(3), pool=3)
    [0, 2, 4]
    
    
    Parameters
    ----------
    fun : function
        *mulit processing* only support`def` in globls() function can't be lambda and inner function 
        which will raise PicklingError: Can't pickle 
    *iterables : list 
        Make an iterator that computes the function using arguments from each of the iterables. 
        Stops when the shortest iterable is exhausted.
    pool : int, default None
        the number of Process or Threading
        the default is the number of CPUs in the system
        进程数目，默认为CPU进程数
    printfreq : int or float, default None
        short of `print frequent`, auto print program progress in `mapmt` and `mapmp`   
        if `printfreq < 1` then `printfreq = len(iterables[0])*printfreq`
        打印进度的频次 默认不打印
    logf : function, default None
        Hook Function For log , every printfreq
        do logf([fun, args, ind, lenn, logf])
    thread : bool, default False
        是否以*多线程*形式替换多进程

    Notice
    ----------
    *  mapmp(fun, *iterables,pool=None, **kv) will raise Error in Python 2.7.13
    * 多进程中 异常无法精确定位
    * numpy操作 由于指令集优化 及numpy版本原因 可能多进程可能会更慢
    * 多进程中 list, dict等元素会被直接复制，无法改变内部元素
    '''    
    if winYl and not kv.get('thread'):
        from boxx import warn1time
        warn1time("""detection your Platform is Windows,   multiprocessing maybe slower cause os.fork is disable.
Even multiprocessing can't work on Windows sometimes""")
    Pool = PoolMp
    if 'thread' in kv and kv['thread']:
        Pool = PoolThread
    pool = Pool(kv['pool']) if 'pool' in kv and kv['pool'] else Pool()
    
    pooln = kv['pool'] if 'pool' in kv else cpun
    logf = kv['logf'] if 'logf' in kv else None
    
    iterables = list(iterables)
    if 'printfreq' in kv and kv['printfreq']:
        printfreq = kv['printfreq']
        l = iterables[-1] = list(iterables[-1])
        lenn = len(l)
        if printfreq < 1:
            printfreq = int(round(lenn*printfreq))
            printfreq = max(1, printfreq)
        def yieldWithIndFun(fun, iterables, lenn, logf, printfreq):
            for i,args in enumerate(zip(*iterables)):
                yield (fun, args, None if i%printfreq else i, lenn, logf)
        chunksize = kv['chunksize'] if 'chunksize' in kv else min(max(1, printfreq*2//3), lenn//pooln)
        re = pool.map(__multiprocessLogFun__, yieldWithIndFun(fun, iterables, lenn, logf, printfreq), chunksize=chunksize)
    else:
        def yieldWithFun(fun, iterables):
            for args in zip(*iterables):
                yield (fun,)+args
        re = pool.map(__multiprocessingFun__,  yieldWithFun(fun, iterables))
    pool.close()
    pool.join()
    return re

def mapmt(fun, *mapArgList, **kv):
    '''
    Map with Multi Threading:多线程版本的map函数， 适用于IO密集型任务
    mapmt(fun, sequence[, sequence, ...], pool=None)->list
    
    >>> mapmt(np.add, range(3), range(3), pool=3)
    [0, 2, 4]
    
    
    Parameters
    ----------
    fun : function
        与 map 使用相同的函数 
    *mapArgList : list 
        用于fun的参数list, fun需要N个参数则有N个列表
    pool : int, default None
        线程数目，默认为CPU进程数
    '''
    return mapmp(fun, *mapArgList,thread=True, **kv)

def maptry(fun, *mapArgList, **kv):
    '''
    test spend time while use `map`, `mapmt`, `mapmp`
    '''
    from boxx import timeit, pblue
    pblue('Begin test map!')
    with timeit('map2'):
        list(map(fun, *mapArgList))
    with timeit('mapmt'):
        list(mapmt(fun, *mapArgList, **kv))
    with timeit('mapmp'):
        list(mapmp(fun, *mapArgList, **kv))


class multiThread():
    '''
    将多线程封装成为类的形式
    
    Example
    ----------
    >>> mt = multiThread(pool=8)
    >>> for arg in args:
            mt(f,*arg)
    >>> mt.join()
    
    Init Parameters
    ----------
    pool : int, default 8
        线程池的最大线程数目
        
    Method
    ----------
    __call__(fun,*l,**kv):
        fun : 需要多线程的函数 
        *l,**kv : 传给函数的参数们
        新建一个线程 并运行fun(*l,**kv)
    join(): 
        等待至所有线程中止。阻塞主线程直至没有子线程运行。 
    '''
    TIME_GAP = .001
    def __init__(self,pool=8):
        self.pool = pool
        self.left = pool
        self.l=[]
    def __call__(self,fun,*l,**kv):
        while not self.left:
            time.sleep(self.TIME_GAP)
        self.left -=1
        def f():
            try:
                r = fun(*l,**kv)
            finally:
                self.left += 1
        self.l.append(setTimeout(f))
        if len(self.l)>100:
            self.l = self.l[-90:]
    f = lambda fun,*l,**kv:fun(*l,**kv) #用于测试单线程的性能
    def join(self):
        while self.left!=self.pool:
            time.sleep(self.TIME_GAP)
#        print('pool',self.pool)

def retry(fun, times=None, exception=Exception, timeGap=0, log=True):
    '''
    重试一个容易raise Exception的无参数函数fun times次数，被捕获的exception 保存在 retry.e 中
    最后返回 fun() 
    
    Parameters
    ----------
    fun : function
        没有参数的过程函数
    times : int, default None
        try 的次数 默认为无限次
    exception : Exception Class, default builtins.Exception
        需要被捕获的异常， 默认捕获所有异常 
    timeGap : float, default 0
        retry 需要等待多少秒时间
    log : bool, default True
        是否打印错误信息    
    Returns
    -------
    fun() : anything
    '''
    count = 0
    while (True if times is None else times - 1):
        try:
            r = fun()
            return r
        except exception as e:
            retry.e = e
            if log:
                fname = fun.__name__ if 'func_name' in dir(fun) else str(fun)
                print((('\x1b[31m%s\x1b[0m'%"%dth's Exception: %s of %s")%(count, typestr(e), fname)))
                print(e)
                print('')
        if timeGap:
            time.sleep(timeGap)
        if times is not None:
            times -= 1    
        count += 1
    return fun()

if __name__ == "__main__":
    
    def fun():
        for i in range(10):
            print(i)
            time.sleep(1)
    from threading import Timer
    thread = Timer(0,fun)
    thread.start()
    pass


# -*- coding: utf-8 -*-
"""
some numpy function

@author: yanglei
"""

from __future__ import unicode_literals

from ..ylsys import sysi
from ..ylcompat import interactivePlot

from ..tool.toolTools import filterList

import numpy as np
from numpy import e, pi, nan, inf
nan = nan * .22
inf = inf * .22
eps = 1e-5

def savenp(path, arr=None):
    '''压缩存储 np.array 为path路径 
    ps: int, bool 压缩效果佳 可达到20倍'''
    if isinstance(path,np.ndarray) and arr is None:
        path,arr = 'savenp_default.npz',path
    np.savez_compressed(path, arr)
def loadnp(path='savenp_default.npz'):
    '''读取path路径下的 .npz 返回 np.array'''
    if path[-4:] != '.npz':
        path += '.npz'
    compress = np.load(path)
    arr = compress[compress.files[0]]
    compress.close()
    return arr 

def mapping_array(array, mapping, strict=True):
    """Element-wise mapping array by a kv mapping
    
    Parameters
    ----------
    array : np.array
        array that content are mapping's key 
    mapping: dict, list, tuple .etc
        a kv mapping struct
    strict : bool, default True
        if strict is False, when k not in mapping, will return k
    """
    ks, idx_array = np.unique(array, return_inverse=True)
    if strict:
        idx_to_v = np.array([mapping[k] for k in ks])
    else:
        idx_to_v = np.array([mapping[k] if k in mapping else k for k in ks])
    mappinged_array = idx_to_v[idx_array.reshape(array.shape)]
    return mappinged_array

@interactivePlot    
def __draw3dSurface(X,Y,Z):    
    import matplotlib.pyplot as plt
    from mpl_toolkits.mplot3d import Axes3D
    from matplotlib import cm
    from matplotlib.ticker import LinearLocator, FormatStrFormatter
    fig = plt.figure()
    ax = fig.gca(projection='3d')
    
    surf = ax.plot_surface(X, Y, Z, rstride=1, cstride=1, cmap=cm.coolwarm, linewidth=0, antialiased=False)
    #画表面,x,y,z坐标， 横向步长，纵向步长，颜色，线宽，是否渐变
    
    #ax.set_zlim(-1.01, 1.01)#坐标系的下边界和上边界
    ax.zaxis.set_major_locator(LinearLocator(10))#设置Z轴标度
    ax.zaxis.set_major_formatter(FormatStrFormatter('%.02f'))#Z轴精度
    fig.colorbar(surf, shrink=0.5, aspect=5)#shrink颜色条伸缩比例（0-1），aspect颜色条宽度（反比例，数值越大宽度越窄）
    
    plt.show()


def plot3dSurface(Z):
    '''
    对二维数组Z 画出3d直方图 
    '''
    m, n = Z.shape
    X = list(range(n))
    Y = list(range(m))
    X, Y = np.meshgrid(X, Y)
    __draw3dSurface(X,Y,Z)

__numpyTypeCache = dict()
def getNumpyType(typee='int'):
    if typee in __numpyTypeCache:
        return __numpyTypeCache[typee]
    finds = eval('('+', '.join(['np.'+mu for mu in filterList(typee, dir(np))])+',)')
    types = [x for x in finds if type(x)==type]
    if typee == "str":
        types = tuple(types) + getNumpyType('unicode')
    types = tuple(types)
    __numpyTypeCache[typee] = types
    return types
def isNumpyType(array, typee='int'):
    '''
    和isinstance一样的用法 判断np.array.dtype 对象对应 [bool, int, float, str]的类
    注意 isNumpyType([bool],int) 为True 
    '''
    npFloatTypes = getNumpyType('float')
    npIntTypes = getNumpyType('int')
    npBoolTypes = getNumpyType('bool')
    npStrTypes = getNumpyType('str')
    if isinstance(typee,tuple):
        return any([isNumpyType(array, t) for t in typee])
    if typee in [bool,'bool']:
        return array.dtype in (npBoolTypes)
    if typee in [int,'int']:
        return array.dtype in (npIntTypes+npBoolTypes)
    if typee in [float,'float']:
        return array.dtype in (npFloatTypes)
    if typee in [str,'str','unicode']:
        return array.dtype in (npStrTypes)
    raise Exception("isNumpyType(array, typee) array must be numpy,"+\
    "typee must be tuple or [bool, int, float, str, unicode] ")

def testNumpyMultiprocessing(n=16, nn=1500):
    '''
    test numpy Multiprocessing performance in your enviroment
    
    Parameters
    ----------
    n : int
        How many times to do np.dot
    nn : int
        the shape of the np.ndarray is (nn, nn)
    '''
    from boxx import timeit, mapmp
    ass = [np.random.rand(nn , nn) for i in range(n)]
    bs = [np.random.rand(nn , nn) for i in range(n)]
    s = 'np.dot%s for %s times'%(str(([nn,nn],[nn,nn])), n )
    name = 'map'
    print('\n\nTesting %s in %s ...'%(s, name))
    with timeit(name):
        l = map(np.dot, ass, bs)
        l = list(l)
    
    name = 'mapmp with pool=%d'%sysi.cpun
    print('\n\nTesting %s in %s ...'%(s, name))
    with timeit(name):
        l = mapmp(np.dot, ass, bs)

if __name__ == '__main__':
    
    
    
    pass
    
    
    



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import numpy as np
import pandas as pd

from ..tool import dynamicWraps, findints, glob,dicto
from ..ylsci import isNumpyType
from ..ylimg import (mapp, labelToColor)

from .ylmlEvalu import Evalu
import imp

def binaryDiff(re,gt,size=0.5,lines=50,bound=False):
    '''
    对二分类问题的结果上色
    False Positive: red mgrid
    False Negative: blue mgrid
    lines: 网格线条的间距
    size:线条多粗 即线条是间距的多少倍
    bound:是否画出边缘
    '''
    re, gt = re > 0.5, gt > 0.5
    rem = np.zeros(gt.shape,int)
    tp = (re) * (gt)
#    tn = (~re) * (~gt)
    fp = re*(~gt)
    fn = (~re)*gt
#    show(tp,tn,fp,fn)
    rem[tp] = 1
    rem[fp] = 2
    rem[fn] = 3
    c=[[0]*3,[1]*3,[1,0,0],[.1,.1,.6]]
    diff = classDiff(rem,re,c,size=size,lines=lines,bound=bound)
    return diff

def drawBoundAndBackground(img,mask,bg=None,replace=False,lines=50,
                           size=0.2,bound=True,boundmode='thick'):
    '''
    给出mask 将给mask区域填充背景色bg的线条 并加上黑白边框
    mask: 所要标注区域
    bg : 背景填充 可以为颜色|图片 默认为红色
    replace : 是否在原图上操作
    lines: 线条的间距
    size:线条多粗 即线条是间距的多少倍
    bound:是否画出边缘
    boundmode: thick 粗正好在边界 'inner'只在前景里面 但是较细
    '''
    assert mask.ndim ==2, 'mask 必须为布尔值'
    if not mask.any():
        return img
    isint = isNumpyType(img,int)
    if not replace:
        img = img.copy()
    if bg is None:
        bg =  [max(255,img.max()),128,0]if isint else [1.,.5,0]
    white = max(255,img.max()) if isint else 1.
    m,n=img.shape[:2]
    i,j = np.mgrid[:m,:n]
    
    step = (m+n)//2//lines
    a = int(step*(1-size))
    drawInd = ~np.where(((i%step<a)& (j%step<a)),True,False)
#    from tool import g
#    g.x = mask,drawInd, bg,img
    if isinstance(bg,np.ndarray) and bg.ndim >=2:
        img[mask*drawInd] = bg[mask*drawInd]
    else:
        img[mask*drawInd] = bg
    if bound:
        from skimage.segmentation import find_boundaries
        boundind = find_boundaries(mask, mode=boundmode,background=True)
        boundBg = np.where((i+j)%10<5,white,0)
        img[boundind] = boundBg[boundind][...,None]
    return (img)
   
def classDiff(rem,gtm,colors=None,size=.15,reMod=False,lines=50,bound=True):
    '''
    对多分类问题的gt进行上色
    对于错误分类 加网格(网格颜色是resoult的颜色) 加有边框
    rem :多分类结果 二维矩阵
    gtm :GroundTruth
    colors:标签对应的颜色
    size:网格所占用的比重
    reMod:对resoult上色, 对于错误分类 网格颜色是 GroundTruth的颜色
    lines:网格线的数量
    bound:是否画出边缘
    '''
    assert rem.ndim==2 and gtm.ndim==2,"rem,gtm 's dim must be 2"
    rgb = labelToColor(rem if reMod else gtm, colors) 
    clas = list(range(len(colors)))
    for c,color in enumerate((colors)): #c mean iter every Class
        for oc in clas: # oc means OtherClass
            if oc==c:
                continue
            mask = (rem==c)*(gtm==oc)
            if mask.any():
                bg = colors[oc if reMod else c]
#                print bg,c,oc
                drawBoundAndBackground(rgb,mask,bg,bound=bound,
                                       size=size,replace=True,lines=lines)
    return rgb

def confusionMatrix(re,gt,classn=None):
    '''求混淆矩阵（confusion matrix）
    
    Parameters
    ----------
    re : np.ndarray
        resoult , 预测的标签
    gt : np.ndarray 
        ground Truth, 值为每个像素的类别
    classn : int,  default None
        总类别数目 默认为 max(re.max(), gt.max())+1
    '''
    if classn is None:
        classn = max(re.max(), gt.max())+1
    ma = np.zeros((classn,classn),int)
    for ind in range(classn):
        row = np.histogram(re[(gt==ind)],classn,range=(0,classn))[0]
        ma[ind][:] = row[:]
    return ma



def f1Score(re,gt,classn):
    '''求各个类别的f1Score 
    先求混淆矩阵（confusion matrix）
    
    Parameters
    ----------
    re : np.ndarray
        resoult or prob, 预测的标签或h*w*n 的概率矩阵
    gt : np.ndarray 
        ground Truth, 值为每个像素的类别
    classn : int
        总类别数目
    '''
    if re.ndim == 3:
        re = re.argmax(2)
    cma = confusionMatrix(re,gt,classn)
    ma = np.float64(cma)
    tp = ma[list(range(classn)),list(range(classn))]
    fp = ma.sum(0)-tp
    fn = ma.sum(1)-tp
    precision = tp/(tp+fp)
    recall = tp/(tp+fn)
    f1 = 2*precision*recall/(precision+recall)
    return f1

@dynamicWraps
def getWeightCore(hh,ww=None,mappFun=None,seta=0.5):
    '''
    返回一个权重二维矩阵 ，默认是seta=0.5的二维高斯分布
    mappFun: 对矩阵每个点执行mappFun(i,j) 返回的结果构成二维矩阵
    '''
    if ww is None:
        ww = hh
    if mappFun is None:
#        ijToCenter = lambda x,i,j:(((i/float(hh)-1/2.)**2+(j/float(ww)-1/2.)**2))
#        wc = weightCore = mapp(ijToCenter,weightCore,need_i_j=True)
        i,j = np.mgrid[:hh,:ww]
        wc = (((i/float(hh)-1/2.)**2+(j/float(ww)-1/2.)**2))
        wc = 1./(2*np.pi*seta**2)*np.e**(-wc/(2*seta**2))
        wc = wc/wc.max()
        #show(normalizing(img[:hh,:ww]*wc[...,None]),img[:hh,:ww])
#        polt3dSurface(wc)
        return wc
    weightCore = np.zeros((hh,ww))
    return mapp(lambda x,i,j:mappFun(i,j),weightCore,need_i_j=True)

def smallImg(img,simgShape, step=None,f=None):
    '''
    将大图切割成固定大小的小图,使产生的小图能覆盖大图的所有面积

    Parameters
    ----------
    simgShape : int or tuple or float
        小图片的shape,为int时候 自动转换为(simgShape, simgShape)
        为float时候 转换为 (int(h*simgShape),int(w*simgShape))
    step : float or int or tuple(steph,stepw),defalut None
        h和w方向上 相邻切割的步长, 默认为simgShape 
        float : (int(step*simgShape[0]),int(step*simgShape[1]))
        int : (step, step)
    fun : funcatin, default None
        若有fun 则执行fun(simg,i,j)
        其中：
            simg:被切割的小图片
            i: simg所在img的row
            j: simg所在img的col
    
    Returns
    -------
    simgs : list of ndarray
        切割出来的小图的list
    '''
    h,w = img.shape[:2]
    if isinstance(simgShape,float):
        hh,ww = (int(h*simgShape),int(w*simgShape))
    if isinstance(simgShape,int):
        hh,ww = simgShape,simgShape
    if isinstance(simgShape,(tuple,list)):
        hh,ww = simgShape
    if step is None:
        steph,stepw = hh,ww
    if isinstance(step,int):
        steph,stepw = step,step
    if isinstance(step,float):
        steph,stepw = int(hh*step),int(ww*step)
    if isinstance(step,(tuple,list)):
        steph,stepw = step
    simgs = []
    for i in list(range(0,h-hh,steph))[:]+[h-hh]:
        for j in list(range(0,w-ww,stepw))[:]+[w-ww]:
            simg = img[i:i+hh,j:j+ww]
            simgs.append(simg)
            if f:
                f(simg,i,j)
    return simgs


def autoSegmentWholeImg(img,simgShape,handleSimg,step=None,weightCore=None):
    '''
    将img分割到 shape为simgShape 的小图simg，执行handleSimg(simg)
    将所有handleSimg(simg)的结果自动拼接成img形状的ndarray并返回
    
    Parameters
    ----------
    img : ndarray
        需要被分割处理的图片
    simgShape : int or tuple
        小图片的shape,为int时候 自动转换为(simgShape, simgShape)
    handleSimg : function
        用于处理shape为simgShape的小图片的函数 
        此函数需要接受一个ndarray作为参数并返回shape[:2]同为为(h,w)的ndarray
        即：handleSimg(simg)=>ndarray，比如 net.pridict(simg)
    step : float or int or tuple(steph,stepw),defalut None
        h和w方向上 相邻切割的步长, 默认为simgShape 
        float : (int(step*simgShape[0]),int(step*simgShape[1]))
        int : (step, step)
    weightCore : {None,'avg','gauss',ndarray}, defalut None 
        对于两个simg图片重叠部分进行融合时候的各自权重
        默认取距离simg中心最近的部分
       'gauss':在重叠部分 采用高斯分布 使之离simg中心越远，权重越低
       'avg':重叠部分取平均
    
    Returns
    -------
    result : ndarray
        shape[:2]等于img.shape[:2]的ndarray
    '''
    if isinstance(simgShape,int):
        hh,ww = simgShape,simgShape
    hh,ww = simgShape
    h,w = img.shape[:2]
    if weightCore is None:
        pass
    elif isinstance(weightCore,np.ndarray):
        pass
    elif weightCore in ['avg']:
        weightCore = np.ones((hh,ww))
    elif weightCore in ['guss','gauss']:
        weightCore = getWeightCore(hh,ww)
    else:
        raise Exception('Illegal argus `weightCore` in `autoSegmentWholeImg`!')
    weight = np.zeros((h,w))
    class c:
        re=None
        near=None
    def f(simg,i,j):
        sre = handleSimg(simg)
        if c.re is None:
            c.re = np.zeros((h,w)+sre.shape[2:],sre.dtype)
        if weightCore is None:
            if c.near is None:
                y,x = np.mgrid[:hh,:ww]
                c.near = 1-((x*1./ww-1./2)**2+(y*1./hh-1./2)**2)**.5
            ind = c.near > weight[i:i+hh,j:j+ww]
            c.re[i:i+hh,j:j+ww][ind]= sre[ind]
            weight[i:i+hh,j:j+ww][ind]= c.near[ind]
            return
        oldw = weight[i:i+hh,j:j+ww]
        ws = weightCore
        if sre.ndim!=2:
            ws = ws[...,None]
            oldw = oldw[...,None]
    #    map(loga,[ws,sre,c.re,oldw,c.re[i:i+hh,j:j+ww]*oldw])
        c.re[i:i+hh,j:j+ww] = (ws*sre + c.re[i:i+hh,j:j+ww]*oldw)/(ws+oldw)
        weight[i:i+hh,j:j+ww] += weightCore
    #    show(c.re,weight)
    (smallImg(img,(hh,ww),step=step,f=f))
    return c.re

class ArgList(list):
    '''
    标记类 用于标记需要被autoFindBestParams函数迭代的参数列表
    '''
    pass

    
def autoFindBestParams(c, args,evaluFun,sortkey=None, savefig=False):
    '''遍历args里面 ArgList的所有参数组合 并通过sortkey 找出最佳参数组合
    
    Parameters
    ----------
    c : dicto
        即configManager 生成的测试集的所有环境配置 c
        包含args，数据配置，各类函数等
    args : dicto
        predict的参数，但需要包含 ArgList 类 将遍历ArgList的所有参数组合 并找出最佳参数组合
    evaluFun : Function
        用于评测的函数，用于Evalu类 需要返回dict对象
    sortkey : str, default None
        用于筛选时候的key 默认为df.columns[-1]
    
    Return: DataFrame
        每个参数组合及其评价的平均值
    '''
    iters = [it for it in list(args.items()) if isinstance(it[1],ArgList)]
    iters = sorted(iters,key=lambda x:len(x[1]),reverse=True)
    argsraw = args.copy()
    argsl = []
    args = dicto()
    
    k,vs = iters[0]
    lenn = len(iters)
    deep = 0
    tags = [0,]*lenn
    while deep>=0:
        vs = iters[deep][1]
        ind = tags[deep]
        if ind != len(vs):
            v = vs[ind]
            tags[deep]+=1
            key = iters[deep][0]
            args[key] = v
            if deep == lenn-1:
                argsl.append(args.copy())
            else:
                deep+=1
        else:
            tags[deep:]=[0]*(lenn-deep)
            deep -= 1
    assert len(argsl),"args don't have ArgList Values!!"
    pds,pddf = pd.Series, pd.DataFrame
    edic={}
    for arg in argsl:
        argsraw.update(arg)
        c.args.update(argsraw)
        e = Evalu(evaluFun,
                  evaluName='tmp',
                  sortkey=sortkey,
                  loged=False,
                  saveResoult=False,
                  )
        if 'reload' in c:
            imp.reload(c.reload)
            inference = c.reload.inference
        elif 'predictInterface' in c:
            imp.reload(c.predictInterface)
            inference = c.predictInterface.predict
        else:
            raise Exception("don't reload get c.reload")
        for name in c.names[::]:
            gt = c.readgt(name)
            prob = inference(c.toimg(name))
            re = prob.argmax(2)
#            from boxx import g
#            g.re,g.gt = re,gt
            e.evalu(re,gt,name)
    #        img = readimg(name)
    #        show(re,gt)
    #        show(img)
        if sortkey is None:
            sortkey = e.columns[-1]
        keys = tuple(arg.values())
        for k,v in list(arg.items()):
            e[k] = v
        edic[keys] = e
        print('arg: %s\n'%str(arg), e.mean())
    es = pddf([pds(x.mean()) for x in list(edic.values())])
    print('-'*20+'\nmax %s:\n'%sortkey,es.loc[es[sortkey].argmax()])
    print('\nmin %s:\n'%sortkey,es.loc[es[sortkey].argmin()])
    if len(iters) == 1:
        k = iters[0][0]
        import matplotlib.pyplot as plt
        df = es.copy()
        df = df.sort_values(k)
        plt.plot(df[k],df[sortkey],'--');plt.plot(df[k],df[sortkey],'rx')
        plt.xlabel(k);plt.ylabel(sortkey);plt.grid()
        if savefig:
            plt.savefig(savefig)
            plt.close()
        else:
            plt.show()    
    return es

def autoFindBestEpoch(c, evaluFun,sortkey=None,epochs=None, savefig=False):
    '''遍历所有epoch的weight  并通过测试集评估项sortkey 找出最佳epoch
    
    Parameters
    ----------
    c : dicto
        即configManager 生成的测试集的所有环境配置 c
        包含args，数据配置，各类函数等
    evaluFun : Function
        用于评测的函数，用于Evalu类 需要返回dict对象
    sortkey : str, default None
        用于筛选时候的key 默认为df.columns[-1]
    
    Return: DataFrame
        每个参数组合及其评价的平均值
    '''
    args = c.args
    if not isinstance(epochs,(tuple,list)) :
        pas = [p[len(args.prefix):] for p in glob(args.prefix+'*') if p[-4:]!='json']
        eps = [len(findints(s)) and findints(s)[-1] for s in pas]
        maxx = len(eps) and max(eps)
        minn = len(eps) and min(eps)
        if isinstance(epochs,int):
            epochs = list(range(minn,maxx))[::epochs]+[maxx]
        else:
            epochs = list(range(minn,maxx+1))
    args['restore'] = ArgList(epochs)
#    print epochs
    df = autoFindBestParams(c, args, evaluFun,sortkey=sortkey,savefig=savefig)
    return df

if __name__ == '__main__':
    pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: DIYer22@github
@mail: ylxx@live.com
Created on Wed Jan  9 23:24:20 2019
"""
import pandas as pd
import numpy as np


class Markdown(pd.DataFrame):
    """
    Markdown class base on DataFrame
    """

    def round(self, decimals=0, **kv):
        self._decimals = decimals
        redf = pd.DataFrame.round(self, decimals, **kv)
        md = Markdown(redf)
        md._decimals = decimals
        return md

    __round__ = round

    def to_md(self, nblankBetweenCell=1, tableMidSymbol="---:"):
        """
        Transfer DataFrame to markdown
        
        Parameters
        ----------
        df : DataFrame
            pandas.DataFrame
        
        nblankBetweenCell : int, default 1
            How many blanks between 2 cells
        tableMidSymbol : str, default "---:"
            "-:", ":-", ":-:" for markdown 
        
        """
        df = self

        blanks = nblankBetweenCell * " "
        betweenCell = blanks + "|" + blanks
        rowFormat = "|%s%s%s|" % (blanks, "%s", blanks)
        ncols = len(df.columns)
        strcols = df.columns.map(str)
        lencols = np.array(strcols.map(len))

        def try_round(v):
            if "_decimals" in self.__dict__:
                try:
                    format_str = "%." + str(self._decimals) + "f"
                    s = format_str % v
                    return s
                except:
                    pass
            return str(v)

        strdf = df.applymap(try_round)
        lendf = np.array(strdf.applymap(len))

        maxLens = np.array(
            list(zip(lendf.max(0), lencols, [len(tableMidSymbol)] * ncols))
        ).max(1)

        vMulBlank = np.vectorize(lambda x: x * " ")

        mdcols = vMulBlank(maxLens - lencols) + strcols
        mddf = vMulBlank(maxLens - lendf) + strdf

        headstr = rowFormat % betweenCell.join(mdcols)

        tableMidStr = rowFormat % betweenCell.join(
            map(lambda x: x + tableMidSymbol, vMulBlank(maxLens - len(tableMidSymbol)))
        )

        bodystr = "\n".join(
            map(lambda kv: rowFormat % betweenCell.join(kv[-1]), mddf.iterrows())
        )

        tablestr = "\n".join([headstr, tableMidStr, bodystr])
        return tablestr

    def __str__(self):
        return self.to_md()

    @staticmethod
    def test():
        df = pd.DataFrame(
            [
                {"a": 1, "b": 1, "d": "str"},
                {"a": 1, "b": 1, "c": 0.1},
                {"a": 1, "b": 1 / 2, "c": 1 / 3},
            ]
        )
        md = Markdown(df).round(2)
        print(md)
        print("-" * 20)
        print(md.round(2))
        return df


if __name__ == "__main__":
    Markdown.test()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
'''
`import boxx.p` act same as `p()`

Notice: For Python 2 only first time import works 

useage:
    >>> import boxx.p
    
        
Instruct
----------
    your can use `out()`, `p()` in any function or module
    exec `out()`, all vars that belong the function will transport to
    Python interactive shell. and `globals()` will in `p` which is a dicto
    
    BTW, `import boxx.out`, `import boxx.p` is convenient way to use `out()` without `from boxx import out`
    
    在函数内运行`p()` or `lc()`  
    函数的 frame等其他信息 则放入全局变量 lc

'''
from .ylsys import  py2
from .tool import p

deep = 6
if py2:
    deep = 1

p(deep)

if py2:
    import sys
    sys.modules['boxx.p'] = p
else:
    from .tool import removeImportSelf
    removeImportSelf('boxx.p')
#from . import *
#print(prettyFrameStack())
#fs = getFatherFrames(endByMain=1)
#list(map(pipe(prettyFrameLocation,log),fs))


if __name__ == '__main__':
    pass




# -*- coding: utf-8 -*-

from __future__ import unicode_literals


from .ylimgTool import sk, np, sda
from .ylimgTool import (show, loga, mapp, normalizing, imsave, imread,
                       standImg, resize)

from ..ylcompat import py2

def base64Img(arr):
    import base64,cv2
    cnt = cv2.imencode('.jpg',arr[:,:,[2,1,0]])[1]
    if not py2:
        return base64.encodebytes(cnt[...,0]).decode('utf-8')
    return base64.encodestring(cnt)

def gifSave(imgs, name='a', fps=None,**kargs):
    '''
    图片序列 保存为GIF
    loop : int
        The number of iterations. Default 0 (meaning loop indefinitely)
    duration : {float, list}
        The duration (in seconds) of each frame. Either specify one value
        that is used for all frames, or one value for each frame.
    fps : float
        The number of frames per second. If duration is not given, the
        duration for each frame is set to 1/fps. Default 10.
    palettesize : int
        The number of colors to quantize the image to. Is rounded to
        the nearest power of two. Default 256.
    '''
    if name.lower()[-4:] != '.gif':
        name += '.gif'
    import imageio
    if fps :
        imageio.mimsave(name, imgs, fps=fps,**kargs)    
    else:
        imageio.mimsave(name, imgs,**kargs)    
        
def videoToImgs(videoPath,begin=0,end=0):
    import cv2
    cap = cv2.VideoCapture(videoPath)  
    if not cap.isOpened():
        raise NameError('video "%s" can\'t read by open CV !'%videoPath)
    number = 0
    frames = []
    while(cap.isOpened()):  
        ret, frame = cap.read()  
        if number>=begin:
            frame = frame[...,[2,1,0]]
            frames.append(frame)
        number+= 1
        if number >= end:
            break
    return frames
    
def generateBigImgForPaper(imgMa,lengh=1980,border=20,saveName='bigImgForPaper.png'):
    '''
    生成科研写作用的样本对比图
    imgMa: 图片行程的二维列表
    lengh: 大图片的宽度, 长度根据imgMa矩阵的高自动算出
    border: 图片与图片间的间隔
    '''
    big = None
    for rr in imgMa:
        rr = list(map(standImg,rr))
        nn = len(rr)
        a = int((lengh-nn*border)/nn)
        m,n = rr[0].shape[:2]
        b = int(a*m/n)
        row = None
        rr = [resize(r,(b,a)) for r in rr]
        for r in rr:

            if row is None:
                row = r
            else:
                row = np.append(row,np.ones((b,border,3)),1)
                row = np.append(row,r,1)
        if big is None:
            big = row
        else:
            big = np.append(big,np.ones((border,big.shape[1],3)),0)
            big = np.append(big,row,0)

    show(big)
    if saveName:
        imsave(saveName,big)
    return big

if __name__ == '__main__':
    pass



# -*- coding: utf-8 -*-
"""
Vector Class

Note:
    we use deg rather than rad, for human

@author: yl
"""
from functools import wraps

import numpy as np
from numpy import pi

from ..tool import strnum
from ..ylcompat import py2

rad2deg = lambda rad: rad / pi * 180
deg2rad = lambda deg: deg / 180 * pi

# TODO mv to ylTriFun.py
def degShift(triFun, arc=False):
    funForInfo = lambda: 0 if py2 else triFun
    @wraps(funForInfo)
    def innerfun(inp):
        if not arc:
            inp = deg2rad(inp)
        r = triFun(inp)
        if arc:
            r = rad2deg(r)
        return r
    return innerfun

sin = degShift(np.sin)
cos = degShift(np.cos)
tan = degShift(np.tan)

arcsin = degShift(np.arcsin, arc=True)
arccos = degShift(np.arccos, arc=True)
arctan = degShift(np.arctan, arc=True)


class Vector(np.ndarray):
    #    def __init__(self, point):
    #        np.ndarray.__init__(self, point)
    #        self._norm = None
    def __new__(self, point):
        obj = np.asarray(point).view(self)
        self._norm = None
        return obj

    @property
    def x(self):
        return self[0]

    @property
    def y(self):
        return self[1]

    @property
    def z(self):
        return self[2]

    @x.setter
    def x(self, v, *l):
        self[0] = v

    @y.setter
    def y(self, v, *l):
        self[1] = v

    @z.setter
    def z(self, v, *l):
        self[2] = v

    @property
    def h(self):
        return self[0]

    @property
    def w(self):
        return self[1]

    @h.setter
    def h(self, v, *l):
        self[0] = v

    @w.setter
    def w(self, v, *l):
        self[1] = v

    def intround(self):
        return self.round().astype(int)

    @property
    def norm(self):
        if self._norm is None:
            self._norm = (np.array(self) ** 2).sum() ** 0.5
        return self._norm

    def angleWith(self, v):
        cosin = (self * v).sum() / (self.norm * v.norm)
        deg = arccos(cosin)
        return deg

    def rotation(self, deg):
        assert self.shape == (2,)
        transferMatrix = [[cos(deg), -sin(deg)], [sin(deg), cos(deg)]]
        after = np.dot(transferMatrix, self.T).T
        return Vector(after)

    def __str__(self,):
        if (self).shape == (2,):
            if "int" in str(self.dtype):
                return "Vector(x|h=%s, y|w=%s)%s" % ((self[0]), (self[1]), self.dtype)
            return "Vector(x|h=%s, y|w=%s)%s" % (
                strnum(self[0]),
                strnum(self[1]),
                self.dtype,
            )
        if (self).shape == (3,):
            if "int" in str(self.dtype):
                return "Vector(x|h=%s, y|w=%s, z=%s)%s" % (
                    (self[0]),
                    (self[1]),
                    (self[2]),
                    self.dtype,
                )
            return "Vector(x|h=%s, y|w=%s, z=%s)%s" % (
                strnum(self[0]),
                strnum(self[1]),
                strnum(self[2]),
                self.dtype,
            )
        return np.ndarray.__str__(self) + " (Vector)"

    __repr__ = __str__

    @staticmethod
    def test():
        a = Vector([4, 0])
        b = Vector([4, 4 * 3 ** 0.5])
        print(a.angleWith(b))
        print(a.rotation(45).norm)


v0 = Vector([0, 0])

if __name__ == "__main__":
    Vector.test()
    pass
